#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import json
import logging
import os
import subprocess

from .data_utils.example import VALID_FEATURE_FIELDS
from .paraphrase.transformers_utils import MODEL_PARALLEL_SUPPORTED_MODELS
from .tasks.registry import get_tasks
from .util import have_multilingual

logger = logging.getLogger(__name__)


def get_commit():
    directory = os.path.dirname(__file__)
    return (
        subprocess.Popen("cd {} && git log | head -n 1".format(directory), shell=True, stdout=subprocess.PIPE)
        .stdout.read()
        .split()[1]
        .decode()
    )


def save_args(args, force_overwrite=False):
    os.makedirs(args.log_dir, exist_ok=args.exist_ok or force_overwrite)
    variables = vars(args).copy()
    # remove the task objects before saving the configuration to the JSON file,
    # because tasks are not JSON serializable.
    del variables['train_tasks']
    del variables['val_tasks']
    with open(os.path.join(args.log_dir, 'config.json'), 'wt') as f:
        json.dump(variables, f, indent=2)


def parse_argv(parser):
    parser.add_argument('--root', default='.', type=str, help='root directory for data, results, embeddings, code, etc.')
    parser.add_argument('--data', default='.data/', type=str, help='where to load data from.')
    parser.add_argument('--save', required=True, type=str, help='where to save results.')
    parser.add_argument('--embeddings', default='.embeddings/', type=str, help='where to save embeddings.')
    parser.add_argument('--cache', default='.cache/', type=str, help='where to save cached files')

    parser.add_argument(
        '--train_languages',
        type=str,
        default='en',
        dest='train_src_languages',
        help='Specify dataset source languages used during training for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )
    parser.add_argument(
        '--eval_languages',
        type=str,
        default='en',
        dest='eval_src_languages',
        help='Specify dataset source languages used during validation for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )

    parser.add_argument(
        '--train_tgt_languages',
        type=str,
        default='en',
        help='Specify dataset target languages used during training for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )
    parser.add_argument(
        '--eval_tgt_languages',
        type=str,
        default='en',
        help='Specify dataset target languages used during validation for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )

    parser.add_argument(
        '--train_tasks', nargs='+', type=str, dest='train_task_names', help='tasks to use for training', required=True
    )
    parser.add_argument('--train_iterations', nargs='+', type=int, help='number of iterations to focus on each task')
    # TODO rename to train_batch_size; keeping it for now for backward compatibility
    parser.add_argument(
        '--train_batch_tokens',
        nargs='+',
        default=[400],
        type=int,
        help='Number of tokens to use for dynamic batching, corresponding to tasks in train tasks.'
        'If sentence_batching is used, this will be interpreted as number of examples.',
    )
    parser.add_argument('--jump_start', default=0, type=int, help='number of iterations to give jump started tasks')
    parser.add_argument('--n_jump_start', default=0, type=int, help='how many tasks to jump start (presented in order)')
    parser.add_argument(
        '--num_print', default=10, type=int, help='how many validation examples with greedy output to print to std out'
    )
    parser.add_argument(
        '--print_train_examples_too',
        action='store_true',
        help='Whether to print some train examples along with eval examples during validation',
    )

    parser.add_argument('--no_tensorboard', action='store_false', dest='tensorboard', help='Turn off tensorboard logging')
    parser.add_argument(
        '--tensorboard_dir', default=None, help='Directory where to save Tensorboard logs (defaults to --save)'
    )
    parser.add_argument('--max_to_keep', default=1, type=int, help='number of checkpoints to keep')
    parser.add_argument('--log_every', default=100, type=int, help='how often to log results in # of iterations')
    parser.add_argument('--save_every', default=1000, type=int, help='how often to save a checkpoint in # of iterations')

    parser.add_argument(
        '--val_tasks', nargs='+', type=str, dest='val_task_names', help='tasks to collect evaluation metrics for'
    )
    parser.add_argument('--val_every', default=1000, type=int, help='how often to run validation in # of iterations')
    parser.add_argument(
        '--val_batch_size',
        nargs='+',
        default=[3000],
        type=int,
        help='Number of tokens in each batch for validation, corresponding to tasks in --val_tasks',
    )

    parser.add_argument(
        '--sentence_batching', action='store_true', help='Batch same sentences together (used for multilingual tasks)'
    )
    parser.add_argument(
        '--use_encoder_loss',
        action='store_true',
        help='Force encoded values for sentences in different languages to be the same',
    )
    parser.add_argument(
        '--encoder_loss_type',
        type=str,
        default='mean',
        choices=['mean', 'sum'],
        help='Function to calculate encoder_loss from the context hidden states',
    )
    parser.add_argument(
        '--encoder_loss_weight',
        type=float,
        default=0.1,
        help='multiplicative constant choosing the weight of encoder_loss in total loss',
    )
    parser.add_argument('--eval_set_name', type=str, help='Evaluation dataset name to use during training')

    parser.add_argument('--max_output_length', default=150, type=int, help='maximum output length for generation')
    parser.add_argument('--max_generative_vocab', default=50000, type=int, help='max vocabulary for the generative softmax')
    parser.add_argument('--subsample', default=20000000, type=int, help='subsample the datasets')
    parser.add_argument('--preserve_case', action='store_false', dest='lower', help='whether to preserve casing for all text')
    parser.add_argument(
        "--reduce_metrics",
        type=str,
        default='max',
        choices=['max'],
        help='How to calculate the metric when there are multiple outputs per input.',
    )

    # These are generation hyperparameters. Each one can be a list of values in which case, we generate `num_outputs` outputs for each set of hyperparameters.
    parser.add_argument("--num_outputs", type=int, nargs='+', default=[1], help='number of sequences to output per input')
    parser.add_argument("--temperature", type=float, nargs='+', default=[0.0], help="temperature of 0 implies greedy sampling")
    parser.add_argument(
        "--repetition_penalty",
        type=float,
        nargs='+',
        default=[1.0],
        help="primarily useful for CTRL model; in that case, use 1.2",
    )
    parser.add_argument("--top_k", type=int, nargs='+', default=[0], help='0 disables top-k filtering')
    parser.add_argument("--top_p", type=float, nargs='+', default=[1.0], help='1.0 disables top-p filtering')
    parser.add_argument("--num_beams", type=int, nargs='+', default=[1], help='1 disables beam seach')
    parser.add_argument("--num_beam_groups", type=int, nargs='+', default=[1], help='1 disables diverse beam seach')
    parser.add_argument("--diversity_penalty", type=float, nargs='+', default=[0.0], help='0 disables diverse beam seach')
    parser.add_argument(
        "--no_repeat_ngram_size",
        type=int,
        nargs='+',
        default=[0],
        help='ngrams of this size cannot be repeated in the output. 0 disables it.',
    )

    parser.add_argument(
        '--model',
        type=str,
        choices=['TransformerLSTM', 'TransformerSeq2Seq', 'TransformerForTokenClassification'],
        default='TransformerLSTM',
        help='which model to import',
    )
    parser.add_argument(
        '--pretrained_model',
        default=None,
        help='which pretrained model to use on the encoder side; choose a name from Huggingface models',
    )

    parser.add_argument(
        '--num_workers', type=int, default=0, help='Number of processes to use for data loading (0 means no multiprocessing)'
    )

    parser.add_argument(
        '--rnn_dimension', default=None, type=int, help='output dimensions for RNN layers (for TransformerLSTM)'
    )
    parser.add_argument('--rnn_layers', default=1, type=int, help='number of layers for RNN modules ')
    parser.add_argument(
        '--rnn_zero_state',
        default='average',
        choices=['zero', 'average', 'cls'],
        help='how to construct RNN zero state (for TransformerLSTM)',
    )

    parser.add_argument(
        '--trainable_decoder_embeddings', default=50, type=int, help='size of decoder embedding (for TransformerLSTM)'
    )
    parser.add_argument('--dropout_ratio', default=0.2, type=float, help='dropout for the model (for TransformerLSTM)')

    parser.add_argument('--override_context', type=str, default=None, help='Override the context for all tasks')
    parser.add_argument('--override_question', type=str, default=None, help='Override the question for all tasks')
    # TODO for backward compatibility only. Remove after no old model (including paraphraser) is in use.
    parser.add_argument(
        '--no_separator',
        action='store_true',
        help='By default, we add a model-specific separator token between question and context when concatenating them. This argument disables that.',
    )
    parser.add_argument(
        "--almond_has_single_program",
        action='store_false',
        dest='almond_has_multiple_programs',
        help='Indicate if almond dataset has multiple programs for each sentence',
    )
    parser.add_argument(
        '--almond_lang_as_question',
        action='store_true',
        help='if true will use "Translate from ${language} to ThingTalk" for question',
    )
    parser.add_argument(
        '--almond_detokenize_sentence',
        action='store_true',
        help='undo word tokenization of almond sentence fields (useful if the tokenizer is sentencepiece)',
    )
    parser.add_argument(
        '--almond_thingtalk_version', type=int, choices=[1, 2], default=2, help='Thingtalk version for almond datasets'
    )

    parser.add_argument('--preprocess_special_tokens', action='store_true', help='convert special ThingTalk tokens to words')

    parser.add_argument(
        '--model_parallel',
        action='store_true',
        help='Use model parallelization by spliting model weights across available gpus',
    )
    parser.add_argument(
        '--mp_device_ratio',
        default=None,
        nargs='+',
        type=int,
        help='Provide the distribution ratio of model layers across gpus when using model_parallel'
        'e.g. "1 2 2 2" first device recieves half number of layers compared to other devices'
        'default is None meaning we distribute evenly on all available gpus',
    )

    parser.add_argument('--warmup', default=40, type=int, help='warmup for learning rate. setting it to 1 disables warmup.')
    parser.add_argument('--grad_clip', default=1.0, type=float, help='gradient clipping')
    parser.add_argument(
        '--beta0',
        default=0.9,
        type=float,
        help='alternative momentum for Adam (only when not using transformer scheduler), and RAdam',
    )
    parser.add_argument(
        '--optimizer', default='adam', choices=['adam', 'adamw', 'sgd', 'radam'], type=str, help='optimizer to use'
    )
    parser.add_argument(
        '--lr_schedule',
        type=str,
        default='transformer',
        choices=['transformer', 'constant', 'linear', 'sgd', 'cosine'],
        help='The learning rate strategy. All of them can be used with or without warmup.',
    )
    parser.add_argument(
        '--lr_multiply',
        default=0.01,
        type=float,
        help='Multiplier for the `transformer` learning rate scheduler, constant value for `constant` and maximum value for `linear` and `cosine` schedulers.',
    )
    parser.add_argument('--weight_decay', default=0.0, type=float, help='weight L2 regularization')
    parser.add_argument(
        '--gradient_accumulation_steps',
        default=1,
        type=int,
        help='Number of accumulation steps. Useful to effectively get larger batch sizes.',
    )

    # Loss Truncation; introduced in https://arxiv.org/abs/2004.14589
    parser.add_argument(
        '--dropper_ratio',
        type=float,
        default=0.0,
        help='Ratio of dropped examples in the "Loss Truncation" algorithm. 0 disables truncation.',
    )
    parser.add_argument(
        '--dropper_min_count',
        type=int,
        default=10000,
        help='Number of examples to see in the "Loss Truncation" algorithm before starting to drop high-loss examples.',
    )
    # Label smoothing; see https://arxiv.org/abs/1906.02629 for detailed analysis on its effect on neural network calibration
    parser.add_argument(
        '--label_smoothing',
        type=float,
        default=0.0,
        help='A number in [0, 1] to be used for label smoothing. 0 disables smoothing.',
    )

    parser.add_argument(
        '--load',
        default=None,
        type=str,
        help='path to checkpoint to load model from inside --args.save, usually set to best.pth',
    )
    parser.add_argument('--resume', action='store_true', help='whether to resume training with past optimizers')

    parser.add_argument('--seed', default=123, type=int, help='Random seed.')
    parser.add_argument('--devices', default=[0], nargs='+', type=int, help='a list of devices that can be used for training')

    parser.add_argument(
        '--no_commit',
        action='store_false',
        dest='commit',
        help='do not track the git commit associated with this training run',
    )
    parser.add_argument(
        '--exist_ok', action='store_true', help='Ok if the save directory already exists, i.e. overwrite is ok'
    )

    parser.add_argument(
        '--no_fast_tokenizer', action='store_true', help='Ignore all conditions and use slow version of huggingface tokenizer'
    )
    parser.add_argument(
        '--force_fast_tokenizer',
        action='store_true',
        help='Ignore all conditions and use fast version of huggingface tokenizer',
    )

    parser.add_argument('--skip_cache', action='store_true', help='whether to use existing cached splits or generate new ones')
    parser.add_argument(
        '--cache_input_data', action='store_true', help='Cache examples from input data for faster subsequent trainings'
    )
    parser.add_argument('--use_curriculum', action='store_true', help='Use curriculum learning')
    parser.add_argument(
        '--aux_dataset', default='', type=str, help='path to auxiliary dataset (ignored if curriculum is not used)'
    )
    parser.add_argument(
        '--curriculum_max_frac', default=1.0, type=float, help='max fraction of harder dataset to keep for curriculum'
    )
    parser.add_argument('--curriculum_rate', default=0.1, type=float, help='growth rate for curriculum')
    parser.add_argument(
        '--curriculum_strategy', default='linear', type=str, choices=['linear', 'exp'], help='growth strategy for curriculum'
    )

    # NED args
    parser.add_argument('--do_ned', action='store_true', help='Collect and use entity features during training')
    parser.add_argument(
        '--database_type', default='json', choices=['json', 'remote-elastic'], help='database to interact with for NER'
    )
    parser.add_argument(
        '--database_dump_typeqid2id',
        action='store_true',
        help='This will create the "type to id" mapping for all entities available in ES database',
    )
    parser.add_argument(
        '--database_dump_canonical2type',
        action='store_true',
        help='This will create the "canonical to type" mapping for all entities available in ES database',
    )

    parser.add_argument(
        '--min_entity_len', type=int, default=1, help='Minimum length for entities when ngrams database_lookup_method is used '
    )
    parser.add_argument(
        '--max_entity_len', type=int, default=4, help='Maximum length for entities when ngrams database_lookup_method is used '
    )
    parser.add_argument(
        '--database_dir',
        type=str,
        default='database/',
        help='Database folder containing all relevant files (e.g. alias2qids, pretrained models for bootleg)',
    )

    parser.add_argument(
        '--bootleg_output_dir',
        type=str,
        default='results_temp',
        help='Path to folder where bootleg prepped files should be saved',
    )
    parser.add_argument('--bootleg_model', type=str, default='bootleg_uncased_mini', help='Bootleg model to use')
    parser.add_argument(
        '--bootleg_prob_threshold',
        type=float,
        default=0.3,
        help='Probability threshold for accepting a candidate for a mention',
    )
    parser.add_argument('--bootleg_post_process_types', action='store_true', help='Postprocess bootleg types')

    parser.add_argument(
        '--entity_type_agg_method',
        choices=['average', 'weighted'],
        default='average',
        help='Method used to aggregate several type embeddings for a single mention',
    )
    parser.add_argument(
        "--entity_word_embeds_dropout",
        default=0.0,
        type=float,
        help='Dropout entity word embeddings with this probability when encoding inputs',
    )

    parser.add_argument(
        "--add_types_to_text",
        default='no',
        choices=['no', 'insert', 'append'],
        help='Method for adding types to input text in text-based NER approach',
    )
    parser.add_argument("--ned_dump_entity_type_pairs", action='store_true', help='Dump entity type pairs')
    parser.add_argument(
        '--ned_retrieve_method',
        default='bootleg',
        choices=['naive', 'entity-oracle', 'type-oracle', 'bootleg'],
        type=str,
        help='how to retrieve types for entities',
    )

    parser.add_argument(
        '--database_lookup_method',
        default='ngrams',
        choices=['ngrams', 'smaller_first', 'longer_first'],
        help='smaller_first: start from one token and grow into longer spans until a match is found,'
        'longer_first: start from the longest span and shrink until a match is found,'
        'ngrams: lookup all ngrams in the text and see if there is a match',
    )

    parser.add_argument('--verbose', action='store_true', help='Print detected types for each token')
    parser.add_argument(
        '--almond_domains', nargs='+', default=[], help='Domains used for almond dataset; e.g. music, books, ...'
    )
    parser.add_argument(
        '--ned_features',
        nargs='+',
        type=str,
        default=[],
        help='Features that will be extracted for each entity. Order is important',
    )
    parser.add_argument(
        '--ned_features_size',
        nargs='+',
        type=int,
        default=[],
        help='Max length of each feature vector. All features are padded up to this length',
    )
    parser.add_argument(
        '--ned_features_default_val', nargs='+', type=float, default=[], help='Default value used for each feature'
    )

    # translation args
    parser.add_argument(
        '--att_pooling',
        type=str,
        default='max',
        help='pooling strategy to calculate cross-attention values across multiple heads',
    )
    parser.add_argument('--plot_heatmaps', action='store_true', help='whether to plot cross-attention heatmaps')
    parser.add_argument(
        '--replace_qp',
        action='store_true',
        help='whether to replace tokens between quotation marks after translation with source values',
    )
    parser.add_argument(
        '--force_replace_qp',
        action='store_true',
        help='if replace_qp is not successful, attempt again by leveraging cross-attention to find text spans',
    )

    # token classification task args
    parser.add_argument('--num_labels', type=int, help='num_labels for classification tasks')
    parser.add_argument('--ner_domains', nargs='+', type=str, help='domains to use for CrossNER task')
    parser.add_argument(
        '--hf_test_overfit',
        action='store_true',
        help='Debugging flag for hf datasets where validation will be performed on train set',
    )


def check_and_update_generation_args(args):
    """
    checks all generation commandline arguments. Since these arguments are all lists and shorthand can be used, we expand them to match the expected length
    for instance, [1.0] becomes [1.0 1.0] if all other generation arguments are of length 2
    """
    hyperparameters = [
        'num_outputs',
        'temperature',
        'top_k',
        'top_p',
        'repetition_penalty',
        'num_beams',
        'num_beam_groups',
        'diversity_penalty',
        'no_repeat_ngram_size',
    ]
    max_hyperparameter_len = max([len(getattr(args, h)) for h in hyperparameters])
    valid_len = [1, max_hyperparameter_len]
    for h in hyperparameters:
        if len(getattr(args, h)) not in valid_len:
            logger.error('Hyperparameters should either have the same number of values as others or have exactly one value.')
        # If only one value is provided, use the same value for all samples
        setattr(args, h, getattr(args, h) * (max_hyperparameter_len // len(getattr(args, h))))

    logger.info('Will output %d sequences for each input.', sum(args.num_outputs))
    return args


def post_parse_general(args):

    for feat in args.ned_features:
        if feat not in VALID_FEATURE_FIELDS:
            raise ValueError(
                'Feature {} is not supported. Please provide valid features from {} list'.format(feat, VALID_FEATURE_FIELDS)
            )

    if args.val_task_names is None:
        args.val_task_names = []
        for t in args.train_task_names:
            if t not in args.val_task_names:
                args.val_task_names.append(t)
    if 'imdb' in args.val_task_names:
        args.val_task_names.remove('imdb')

    args.timestamp = datetime.datetime.now(tz=datetime.timezone.utc).strftime('%D-%H:%M:%S %Z')

    # TODO relax the following assertions by dropping samples from batches in Iterator
    if args.sentence_batching and args.train_batch_tokens[0] % len(args.train_src_languages.split('+')) != 0:
        raise ValueError(
            'Your train_batch_size should be divisible by number of train_src_languages when using sentence batching.'
        )
    if args.sentence_batching and args.val_batch_size[0] % len(args.eval_src_languages.split('+')) != 0:
        raise ValueError(
            'Your val_batch_size should be divisible by number of eval_src_languages when using sentence batching.'
        )

    if len(args.ned_features) != len(args.ned_features_size):
        raise ValueError('You should specify max feature size for each feature you provided')

    if len(args.train_task_names) > 1:
        if args.train_iterations is None:
            args.train_iterations = [1]
        if len(args.train_iterations) < len(args.train_task_names):
            args.train_iterations = len(args.train_task_names) * args.train_iterations
        if len(args.train_batch_tokens) < len(args.train_task_names):

            args.train_batch_tokens = len(args.train_task_names) * args.train_batch_tokens

    if len(args.val_batch_size) < len(args.val_task_names):
        args.val_batch_size = len(args.val_task_names) * args.val_batch_size

    # postprocess arguments
    if args.commit:
        args.commit = get_commit()
    else:
        args.commit = ''

    if have_multilingual(args.train_task_names) and (args.train_src_languages is None or args.eval_src_languages is None):
        raise ValueError('You have to define training and evaluation languages when you have a multilingual task')

    args.log_dir = args.save
    args.dist_sync_file = os.path.join(args.log_dir, 'distributed_sync_file')

    for x in ['data', 'save', 'log_dir', 'dist_sync_file']:
        setattr(args, x, os.path.join(args.root, getattr(args, x)))

    args.num_features = len(args.ned_features)

    # tasks with the same name share the same task object
    train_tasks_dict = get_tasks(args.train_task_names, args)
    args.train_tasks = list(train_tasks_dict.values())
    val_task_dict = get_tasks(args.val_task_names, args, available_tasks=train_tasks_dict)
    args.val_tasks = list(val_task_dict.values())

    save_args(args)

    return args


def post_parse_train_specific(args):
    if args.no_fast_tokenizer and args.force_fast_tokenizer:
        raise ValueError('Both no_fast_tokenizer and force_fast_tokenizer flags are on')

    # TODO relax this assertion by allowing training on multiple languages
    if 'mbart' in args.pretrained_model:
        if len(args.train_src_languages.split('+')) != 1 or set(args.train_src_languages.split('+')) != set(
            args.eval_src_languages.split('+')
        ):
            raise ValueError('For now we only support single language training and evaluation with mbart models')

    if args.model_parallel:
        if args.model == 'TransformerLSTM':
            raise ValueError('Model parallel is not supported for TransformerLSTM models')
        elif args.model == 'TransformerSeq2Seq' and args.pretrained_model not in MODEL_PARALLEL_SUPPORTED_MODELS:
            raise ValueError('Only the following models have model_parallel support: ', MODEL_PARALLEL_SUPPORTED_MODELS)

    if args.mp_device_ratio is not None:
        if len(args.mp_device_ratio) != len(args.devices):
            raise ValueError('When using model_parallel number of provided devices must match the number of mp_device_ratio')

    if args.warmup < 1:
        raise ValueError('Warmup should be a positive integer.')

    if args.use_encoder_loss and not (args.sentence_batching and len(args.train_src_languages.split('+')) > 1):
        raise ValueError('To use encoder loss you must use sentence batching and use more than one language during training.')

    if not (len(args.ned_features) == len(args.ned_features_size) == len(args.ned_features_default_val)):
        raise ValueError('You should specify size and default value for each feature you provided')

    if args.preprocess_special_tokens and args.model == 'TransformerLSTM':
        raise ValueError('Preprocessing special tokens should not be used for TransformerLSTM models')

    if args.model == 'TransformerLSTM' and 'uncased' in args.pretrained_model:
        raise ValueError(
            'You should use the cased version of provided model when not preprocessing special tokens.'
            ' Otherwise the program (answer) will not be tokenized properly'
        )

    args.log_dir = args.save
    if args.tensorboard_dir is None:
        args.tensorboard_dir = args.log_dir

    for x in ['embeddings']:
        setattr(args, x, os.path.join(args.root, getattr(args, x)))

    args.num_features = len(args.ned_features)

    save_args(args, force_overwrite=True)

    args = check_and_update_generation_args(args)
    return args



# Copyright 2020 The Board of Trustees of the Leland Stanford Junior University
#
# Author: Giovanni Campagna <gcampagn@cs.stanford.edu>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch
from torch.nn.parallel.scatter_gather import scatter as orig_scatter

# The following code was copied from pytorch
# Copyright 2016-2019 The PyTorch Authors
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
#    and IDIAP Research Institute nor the names of its contributors may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


def scatter(inputs, target_gpus, dim=0):
    r"""
    Slices tensors into approximately equal chunks and
    distributes them across given GPUs. Duplicates
    references to objects that are not tensors.
    """

    def scatter_map(obj):
        if isinstance(obj, torch.Tensor):
            return orig_scatter(obj, target_gpus, dim=dim)
        if isinstance(obj, tuple) and hasattr(obj, '_fields'):  # namedtuple
            return [type(obj)(*el) for el in zip(*map(scatter_map, obj))]
        if isinstance(obj, tuple) and len(obj) > 0:
            return list(zip(*map(scatter_map, obj)))
        if isinstance(obj, list) and len(obj) > 0:
            return list(map(list, zip(*map(scatter_map, obj))))
        if isinstance(obj, dict) and len(obj) > 0:
            return list(map(type(obj), zip(*map(scatter_map, obj.items()))))
        return [obj for _ in target_gpus]

    # After scatter_map is called, a scatter_map cell will exist. This cell
    # has a reference to the actual function scatter_map, which has references
    # to a closure that has a reference to the scatter_map cell (because the
    # fn is recursive). To avoid this reference cycle, we set the function to
    # None, clearing the cell
    try:
        res = scatter_map(inputs)
    finally:
        scatter_map = None
    return res


def scatter_kwargs(inputs, kwargs, target_gpus, dim=0):
    r"""Scatter with support for kwargs dictionary"""
    inputs = scatter(inputs, target_gpus, dim) if inputs else []
    kwargs = scatter(kwargs, target_gpus, dim) if kwargs else []
    if len(inputs) < len(kwargs):
        inputs.extend([() for _ in range(len(kwargs) - len(inputs))])
    elif len(kwargs) < len(inputs):
        kwargs.extend([{} for _ in range(len(inputs) - len(kwargs))])

    # handle cases where len(inputs) < len(num_gpus) and len(kwargs) > 0
    num_empty = sum([int(len(input) == 0) for input in inputs])
    num_full = len(inputs) - num_empty
    if num_full > 0 and num_empty > 0:
        kwargs = kwargs[:num_full]
        inputs = inputs[:num_full]

    inputs = tuple(inputs)
    kwargs = tuple(kwargs)
    return inputs, kwargs


class NamedTupleCompatibleDataParallel(torch.nn.DataParallel):
    def scatter(self, inputs, kwargs, device_ids):
        return scatter_kwargs(inputs, kwargs, device_ids, dim=self.dim)



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import numpy as np


def parse_argv(parser):

    parser.add_argument('--input_file', type=str)
    parser.add_argument('--output_file', type=str)
    parser.add_argument('--filtering_metric', type=str, default='constant', choices=['mean', 'mean+std', 'all', 'constant'])
    parser.add_argument(
        '--filtering_threshold',
        type=float,
        help='STS threshold score used to filter sentences if filtering_metric is constant',
    )


def main(args):

    all_scores = []

    with open(args.input_file, 'r') as fin:
        for line in fin:
            parts = list(map(lambda p: p.strip(), line.split('\t')))
            id_, orig_sent, para_sent, score, program = parts
            all_scores.append(score)

    all_scores = np.array(all_scores, dtype=float)
    scores_mean = np.mean(all_scores)
    scoers_std = np.std(all_scores)

    if args.filtering_metric == 'mean':
        accepted_ids = all_scores >= scores_mean
    elif args.filtering_metric == 'mean+std':
        accepted_ids = all_scores >= (scores_mean + scoers_std)
    elif args.filtering_metric == 'constant':
        assert args.filtering_threshold is not None
        accepted_ids = all_scores >= args.filtering_threshold
    # accept all
    else:
        accepted_ids = all_scores >= 0.0

    with open(args.input_file, 'r') as fin, open(args.output_file, 'w') as fout:
        for i, line in enumerate(fin):
            if accepted_ids[i]:
                parts = list(map(lambda p: p.strip(), line.split('\t')))
                id_, orig_sent, para_sent, score, program = parts
                fout.write('\t'.join(['P' + id_, para_sent, program]) + '\n')



# Parts of this file were adopted from https://github.com/huggingface/transformers.
# See the original copyright notice below.

# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This script is used for fine-tuning library models on a dataset.
GPT and GPT-2 are fine-tuned using a Causal Language Modeling (CLM) loss while BERT family models are fine-tuned
using a Masked Language Modeling (MLM) loss. BART, MBART, and MARIAN are fine-tuned on supervised data using Cross Entropy (CE) loss.
"""

from __future__ import absolute_import, division, print_function

import glob
import logging
import os
import shutil

import torch
from tensorboardX import SummaryWriter
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from torch.utils.data.distributed import DistributedSampler
from transformers import (
    WEIGHTS_NAME,
    AdamW,
    BartConfig,
    BartForConditionalGeneration,
    BartTokenizer,
    BertConfig,
    BertForMaskedLM,
    BertTokenizer,
    CamembertConfig,
    CamembertForMaskedLM,
    CamembertTokenizer,
    DistilBertConfig,
    DistilBertForMaskedLM,
    DistilBertTokenizer,
    GPT2Config,
    GPT2LMHeadModel,
    GPT2Tokenizer,
    MarianConfig,
    MarianMTModel,
    MarianTokenizer,
    MBartConfig,
    MBartForConditionalGeneration,
    OpenAIGPTConfig,
    OpenAIGPTLMHeadModel,
    OpenAIGPTTokenizer,
    RobertaConfig,
    RobertaForMaskedLM,
    RobertaTokenizer,
    get_linear_schedule_with_warmup,
)

from ..data_utils.progbar import prange, progress_bar
from ..util import set_seed, split_file_on_disk
from .data_utils import add_special_tokens, mask_tokens
from .dataset import LengthSortedSampler, TextDataset
from .model_utils import (
    _rotate_checkpoints,
    check_args,
    freeze_embeds,
    freeze_params,
    get_transformer_schedule_with_warmup,
    shift_tokens_right,
)
from .transformers_utils import GenieMBartTokenizer

logger = logging.getLogger(__name__)


MODEL_CLASSES = {
    'gpt2': (GPT2Config, GPT2LMHeadModel, GPT2Tokenizer),
    'openai-gpt': (OpenAIGPTConfig, OpenAIGPTLMHeadModel, OpenAIGPTTokenizer),
    'bert': (BertConfig, BertForMaskedLM, BertTokenizer),
    'roberta': (RobertaConfig, RobertaForMaskedLM, RobertaTokenizer),
    'distilbert': (DistilBertConfig, DistilBertForMaskedLM, DistilBertTokenizer),
    'camembert': (CamembertConfig, CamembertForMaskedLM, CamembertTokenizer),
    'bart': (BartConfig, BartForConditionalGeneration, BartTokenizer),
    'mbart': (MBartConfig, MBartForConditionalGeneration, GenieMBartTokenizer),
    'marian': (MarianConfig, MarianMTModel, MarianTokenizer),
}


def train(
    args,
    train_dataset,
    model,
    tokenizer,
    input_file_name=None,
    multiple_shards=False,
    init_global_step=0,
    init_epochs_trained=0,
):
    """Train the model"""
    if args.local_rank in [-1, 0]:
        tb_writer = SummaryWriter(logdir=args.tensorboard_dir)

    args.train_batch_size = args.per_gpu_train_batch_size * max(1, args.n_gpu)
    if args.sort_by_length:
        train_sampler = LengthSortedSampler(
            train_dataset, batch_size=args.train_batch_size * args.gradient_accumulation_steps, shuffle=True
        )
    else:
        train_sampler = RandomSampler(train_dataset) if args.local_rank == -1 else DistributedSampler(train_dataset)
    train_dataloader = DataLoader(
        train_dataset, sampler=train_sampler, batch_size=args.train_batch_size, collate_fn=train_dataset.collate_fn
    )

    if args.max_steps > 0:
        t_total = args.max_steps
        args.num_train_epochs = args.max_steps // (len(train_dataloader) // args.gradient_accumulation_steps) + 1
    else:
        t_total = len(train_dataloader) // args.gradient_accumulation_steps * args.num_train_epochs

    # Prepare optimizer and schedule (linear warmup and decay)
    no_decay = ['bias', 'LayerNorm.weight']
    optimizer_grouped_parameters = [
        {
            'params': [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
            'weight_decay': args.weight_decay,
        },
        {'params': [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)], 'weight_decay': 0.0},
    ]
    optimizer = AdamW(optimizer_grouped_parameters, lr=args.learning_rate, eps=args.adam_epsilon)

    if args.scheduler == 'linear':
        scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps=args.warmup_steps, num_training_steps=t_total)
    elif args.scheduler == 'transformer':
        if args.model_type == 'bert':
            dimension = model.config.hidden_size
        elif args.model_type == 'gpt2':
            dimension = model.config.n_embd
        else:
            logger.error('Cannot detect hidden size dimensions in this model type. Config: %s', model.config)
        scheduler = get_transformer_schedule_with_warmup(
            optimizer, num_warmup_steps=args.warmup_steps, num_training_steps=t_total, dimension=dimension
        )
    else:
        logger.error('Unknown scheduler type.')

    # Check if saved optimizer or scheduler states exist
    if os.path.isfile(os.path.join(args.model_name_or_path, 'optimizer.pt')) and os.path.isfile(
        os.path.join(args.model_name_or_path, 'scheduler.pt')
    ):
        # Load in optimizer and scheduler states
        optimizer.load_state_dict(torch.load(os.path.join(args.model_name_or_path, 'optimizer.pt')))
        scheduler.load_state_dict(torch.load(os.path.join(args.model_name_or_path, 'scheduler.pt')))

    if args.fp16:
        try:
            from apex import amp
        except ImportError:
            raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use fp16 training.")
        model, optimizer = amp.initialize(model, optimizer, opt_level=args.fp16_opt_level)

    # multi-gpu training (should be after apex fp16 initialization)
    if args.n_gpu > 1:
        model = torch.nn.DataParallel(model)

    # Distributed training (should be after apex fp16 initialization)
    if args.local_rank != -1:
        model = torch.nn.parallel.DistributedDataParallel(
            model, device_ids=[args.local_rank], output_device=args.local_rank, find_unused_parameters=True
        )

    # Train!
    logger.info("***** Running training *****")
    logger.info("  Input file name = %s", input_file_name)
    logger.info("  Num examples = %d", len(train_dataset))
    logger.info("  Num Epochs = %d", args.num_train_epochs)
    logger.info("  Instantaneous batch size per GPU = %d", args.per_gpu_train_batch_size)
    logger.info(
        "  Total train batch size (w. parallel, distributed & accumulation) = %d",
        args.train_batch_size
        * args.gradient_accumulation_steps
        * (torch.distributed.get_world_size() if args.local_rank != -1 else 1),
    )
    logger.info("  Gradient Accumulation steps = %d", args.gradient_accumulation_steps)
    logger.info("  Total optimization steps = %d", t_total)

    global_step = init_global_step
    epochs_trained = init_epochs_trained
    steps_trained_in_current_epoch = 0
    # Check if continuing training from a checkpoint
    if os.path.exists(args.model_name_or_path):
        # set global_step to gobal_step of last saved checkpoint from model path
        global_step = int(args.model_name_or_path.split('-')[-1].split('/')[0])
        epochs_trained = global_step // (len(train_dataloader) // args.gradient_accumulation_steps)
        steps_trained_in_current_epoch = (
            global_step % (len(train_dataloader) // args.gradient_accumulation_steps)
        ) * args.gradient_accumulation_steps

        logger.info("  Continuing training from checkpoint, will skip to saved global_step")
        logger.info("  Continuing training from epoch %d", epochs_trained)
        logger.info("  Continuing training from global step %d", global_step)
        logger.info("  Will skip the first %d steps in the first epoch", steps_trained_in_current_epoch)

    tr_loss, logging_loss = 0, 0

    model_to_resize = model.module if hasattr(model, 'module') else model  # Take care of distributed/parallel training
    model_to_resize.resize_token_embeddings(len(tokenizer))

    model.zero_grad()
    if multiple_shards:
        train_iterator = prange(epochs_trained, 1, desc="Epoch", disable=args.local_rank not in [-1, 0])
    else:
        train_iterator = prange(
            epochs_trained, int(args.num_train_epochs), desc="Epoch", disable=args.local_rank not in [-1, 0]
        )
    set_seed(args)  # Added here for reproducibility (even between python 2 and 3)
    best_eval_perplexity = float('Inf')
    for _ in train_iterator:
        if args.max_steps > 0 and not multiple_shards:
            total_steps = args.max_steps * args.gradient_accumulation_steps
        else:
            total_steps = len(train_dataloader)
        epoch_iterator = progress_bar(
            train_dataloader, desc="Iteration", disable=args.local_rank not in [-1, 0], total=total_steps
        )
        for step, batch in enumerate(epoch_iterator):

            # Skip past any already trained steps if resuming training
            if steps_trained_in_current_epoch > 0:
                steps_trained_in_current_epoch -= 1
                continue

            inputs, attention_mask, labels, position_ids, segment_ids = batch

            if args.mlm:
                inputs, labels = mask_tokens(inputs, labels, tokenizer, args.mlm_probability, args.mlm_ignore_index)
            inputs = inputs.to(args.device)
            labels = labels.to(args.device)
            attention_mask = attention_mask.to(args.device)
            position_ids = position_ids.to(args.device)
            segment_ids = segment_ids.to(args.device)
            model.train()

            model_inputs = {'input_ids': inputs, 'use_cache': False}

            # prepare inputs for mbart, and marian
            if args.model_type in ['mbart', 'marian']:
                model_inputs['attention_mask'] = attention_mask
                decoder_input_ids = shift_tokens_right(labels, args.mlm_ignore_index)
                decoder_input_ids[decoder_input_ids == args.mlm_ignore_index] = tokenizer.pad_token_id
                model_inputs['decoder_input_ids'] = decoder_input_ids
            elif args.model_type == 'bart':
                # TODO according to huggingface bart should also use shift_tokens_right
                # check if that change affects results
                model_inputs['attention_mask'] = attention_mask
                decoder_input_ids = labels.contiguous()
                decoder_input_ids[decoder_input_ids == args.mlm_ignore_index] = tokenizer.pad_token_id
                model_inputs['decoder_input_ids'] = decoder_input_ids
            else:
                model_inputs.update({'position_ids': position_ids, 'token_type_ids': segment_ids})

            outputs = model(**model_inputs)
            lm_logits = outputs.logits.contiguous()
            assert lm_logits.shape[-1] == model.config.vocab_size

            # CrossEntropyLoss ignore_index defaults to -100
            # If a different mlm_ignore_index is provided we make sure it is ignored when calculating the loss
            ce_loss_fct = torch.nn.CrossEntropyLoss(ignore_index=args.mlm_ignore_index)
            loss = ce_loss_fct(lm_logits.view(-1, lm_logits.shape[-1]), labels.view(-1))

            if args.n_gpu > 1:
                loss = loss.mean()  # mean() to average on multi-gpu parallel training
            if args.gradient_accumulation_steps > 1:
                loss = loss / args.gradient_accumulation_steps

            if args.fp16:
                with amp.scale_loss(loss, optimizer) as scaled_loss:
                    scaled_loss.backward()
            else:
                loss.backward()

            tr_loss += loss.item()
            if (step + 1) % args.gradient_accumulation_steps == 0:
                if args.fp16:
                    torch.nn.utils.clip_grad_norm_(amp.master_params(optimizer), args.max_grad_norm)
                else:
                    torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)
                optimizer.step()
                scheduler.step()  # Update learning rate schedule
                model.zero_grad()
                global_step += 1

            if args.local_rank in [-1, 0] and (
                (args.logging_steps > 0 and global_step % args.logging_steps == 0 and global_step != 0)
                or step == total_steps - 1
            ):
                # Log metrics
                if (
                    args.local_rank == -1 and args.evaluate_during_training
                ):  # Only evaluate when single GPU otherwise metrics may not average well
                    results = evaluate(args, model, tokenizer)
                    if args.aux_eval_data_file is not None:
                        aux_results = evaluate(args, model, tokenizer, aux=True)
                        for key, value in aux_results.items():
                            tb_writer.add_scalar('auxiliary_eval_{}'.format(key), value, global_step)
                    if best_eval_perplexity > results['perplexity']:
                        best_eval_perplexity = results['perplexity']
                        if not os.path.exists(args.output_dir):
                            os.makedirs(args.output_dir)
                        logger.info("Saving new best model to %s", args.output_dir)
                        # Save a trained model, configuration and tokenizer using `save_pretrained()`.
                        # They can then be reloaded using `from_pretrained()`
                        model_to_save = (
                            model.module if hasattr(model, 'module') else model
                        )  # Take care of distributed/parallel training
                        model_to_save.save_pretrained(args.output_dir)
                        tokenizer.save_pretrained(args.output_dir)

                        # Good practice: save your training arguments together with the trained model
                        torch.save(args, os.path.join(args.output_dir, 'training_args.bin'))

                    for key, value in results.items():
                        tb_writer.add_scalar('eval_{}'.format(key), value, global_step)
                # TODO add generated text to tensorboard
                # tb_writer.add_text('eval/generated_text', gen_text, global_step)
                tb_writer.add_scalar('lr', scheduler.get_lr()[0], global_step)
                tb_writer.add_scalar('loss', (tr_loss - logging_loss) / args.logging_steps, global_step)
                logging_loss = tr_loss

            if (
                args.local_rank in [-1, 0]
                and args.save_steps > 0
                and global_step % args.save_steps == 0
                and global_step != 0
                and args.save_total_limit > 0
            ):
                checkpoint_prefix = 'checkpoint'
                # Save model checkpoint
                output_dir = os.path.join(args.output_dir, '{}-{}'.format(checkpoint_prefix, global_step))
                if not os.path.exists(output_dir):
                    os.makedirs(output_dir)
                model_to_save = (
                    model.module if hasattr(model, 'module') else model
                )  # Take care of distributed/parallel training
                model_to_save.save_pretrained(output_dir)
                tokenizer.save_pretrained(output_dir)

                torch.save(args, os.path.join(output_dir, 'training_args.bin'))
                logger.info("Saving model checkpoint to %s", output_dir)

                _rotate_checkpoints(args, checkpoint_prefix)

                torch.save(optimizer.state_dict(), os.path.join(output_dir, 'optimizer.pt'))
                torch.save(scheduler.state_dict(), os.path.join(output_dir, 'scheduler.pt'))
                logger.info("Saving optimizer and scheduler states to %s", output_dir)

            if args.max_steps > 0 and global_step > args.max_steps:
                epoch_iterator.close()
                break
        if args.max_steps > 0 and global_step > args.max_steps:
            train_iterator.close()
            break

    if args.local_rank in [-1, 0]:
        tb_writer.close()

    return global_step, tr_loss


def evaluate(args, model, tokenizer, prefix="", aux=False):
    # Loop to handle MNLI double evaluation (matched, mis-matched)
    eval_output_dir = args.output_dir

    if aux:
        file_path = args.aux_eval_data_file
    else:
        file_path = args.eval_data_file

    eval_dataset = TextDataset(tokenizer, args, file_path=file_path, block_size=args.block_size, evaluate=True)

    if not os.path.exists(eval_output_dir) and args.local_rank in [-1, 0]:
        os.makedirs(eval_output_dir)

    args.eval_batch_size = args.per_gpu_eval_batch_size * max(1, args.n_gpu)
    # Note that DistributedSampler samples randomly
    if args.sort_by_length:
        eval_sampler = LengthSortedSampler(eval_dataset, batch_size=args.eval_batch_size, shuffle=False)
    else:
        eval_sampler = SequentialSampler(eval_dataset)
    eval_dataloader = DataLoader(
        eval_dataset, sampler=eval_sampler, batch_size=args.eval_batch_size, collate_fn=eval_dataset.collate_fn
    )

    # multi-gpu evaluate
    if args.n_gpu > 1:
        model = torch.nn.DataParallel(model)

    # Eval!
    logger.info("***** Running evaluation {} *****".format(prefix))
    logger.info("  Num examples = %d", len(eval_dataset))
    logger.info("  Batch size = %d", args.eval_batch_size)
    eval_loss = 0.0
    nb_eval_steps = 0
    model.eval()

    for batch in progress_bar(eval_dataloader, desc="Evaluating"):
        inputs, attention_mask, labels, position_ids, segment_ids = batch
        if args.mlm:
            inputs, labels = mask_tokens(inputs, labels, tokenizer, args.mlm_probability, args.mlm_ignore_index)
        inputs = inputs.to(args.device)
        labels = labels.to(args.device)
        attention_mask = attention_mask.to(args.device)
        position_ids = position_ids.to(args.device)
        segment_ids = segment_ids.to(args.device)

        with torch.no_grad():
            model_inputs = {'input_ids': inputs, 'use_cache': False}

            if args.model_type in ['mbart', 'marian']:
                model_inputs['attention_mask'] = attention_mask
                decoder_input_ids = shift_tokens_right(labels, args.mlm_ignore_index)
                decoder_input_ids[decoder_input_ids == args.mlm_ignore_index] = tokenizer.pad_token_id
                model_inputs['decoder_input_ids'] = decoder_input_ids
            elif args.model_type == 'bart':
                model_inputs['attention_mask'] = attention_mask
                decoder_input_ids = labels.contiguous()
                decoder_input_ids[decoder_input_ids == args.mlm_ignore_index] = tokenizer.pad_token_id
                model_inputs['decoder_input_ids'] = decoder_input_ids
            else:
                model_inputs.update({'position_ids': position_ids, 'token_type_ids': segment_ids})

            outputs = model(**model_inputs)
            lm_logits = outputs.logits

            # debugging
            if args.debug:
                print()
                print([tokenizer.decode(t, skip_special_tokens=True) for t in lm_logits.max(-1)[1]])

            assert lm_logits.shape[-1] == model.config.vocab_size

            tokenizer.batch_decode(lm_logits.max(-1)[1])

            # Same behavior as modeling_bart.py, besides ignoring pad_token_id
            ce_loss_fct = torch.nn.CrossEntropyLoss(ignore_index=args.mlm_ignore_index)
            loss = ce_loss_fct(lm_logits.view(-1, lm_logits.shape[-1]), labels.view(-1))
            eval_loss += loss.mean().item()

        nb_eval_steps += 1

    eval_loss = eval_loss / nb_eval_steps
    perplexity = torch.exp(torch.tensor(eval_loss))

    result = {"perplexity": perplexity}

    output_eval_file = os.path.join(eval_output_dir, prefix, "eval_results.txt")
    with open(output_eval_file, "a") as writer:
        logger.info("***** Eval results {} *****".format(prefix))
        for key in sorted(result.keys()):
            logger.info("  %s = %s", key, str(result[key]))
            writer.write("%s = %s\n" % (key, str(result[key])))

    return result


def parse_argv(parser):
    ## Required parameters
    parser.add_argument(
        "--output_dir",
        default=None,
        type=str,
        required=True,
        help="The output directory where the model predictions and checkpoints will be written.",
    )

    ## Other parameters
    parser.add_argument(
        "--tensorboard_dir", default=None, type=str, help="The output directory where the tensorboard files will be written."
    )
    parser.add_argument("--train_data_file", default=None, type=str, help="The input training data file.")
    parser.add_argument(
        "--aux_train_data_file", default=None, type=str, help="An input training data file for the target domain."
    )
    parser.add_argument(
        '--start_special_token', type=str, default='<paraphrase>', help='The special token for the start of paraphrases.'
    )
    parser.add_argument(
        '--end_special_token', type=str, default='</paraphrase>', help='The special token for the end of paraphrases.'
    )
    parser.add_argument('--pad_token', type=str, default='<pad>', help='The special token for padding..')
    parser.add_argument(
        '--train_all_tokens',
        action='store_true',
        help='If True, the model will be trained on input and output sequences, as opposed to only tokens of the output sequence',
    )
    parser.add_argument(
        "--reverse_position_ids",
        action='store_true',
        help='If we assume we know the length of the output sequence beforehand, we can do a better job at generation.',
    )
    parser.add_argument('--subsample', default=20000000, type=int, help='subsample the datasets')
    parser.add_argument(
        "--eval_data_file",
        default=None,
        type=str,
        help="An optional input evaluation data file to evaluate the perplexity on (a text file).",
    )
    parser.add_argument(
        "--aux_eval_data_file",
        default=None,
        type=str,
        help="An additional input evaluation data file to evaluate the perplexity on (a text file).",
    )

    parser.add_argument("--model_type", default="bert", type=str, help="The model architecture to be fine-tuned.")
    parser.add_argument(
        "--model_name_or_path", default="bert-base-cased", type=str, help="The model checkpoint for weights initialization."
    )

    parser.add_argument(
        "--mlm", action='store_true', help="Train with masked-language modeling loss instead of language modeling."
    )
    parser.add_argument(
        "--mlm_probability", type=float, default=0.15, help="Ratio of tokens to mask for masked language modeling loss"
    )
    parser.add_argument(
        "--mlm_ignore_index",
        type=int,
        default=-100,
        help="Tokens with this label will be ignore when calculating masked language loss",
    )

    parser.add_argument(
        "--config_name",
        default="",
        type=str,
        help="Optional pretrained config name or path if not the same as model_name_or_path",
    )
    parser.add_argument(
        "--tokenizer_name",
        default="",
        type=str,
        help="Optional pretrained tokenizer name or path if not the same as model_name_or_path",
    )
    parser.add_argument(
        "--cache_dir",
        default=".embeddings",
        type=str,
        help="Optional directory to store the pre-trained models downloaded from s3 (instread of the default one)",
    )
    parser.add_argument(
        "--block_size",
        default=-1,
        type=int,
        help="Optional input sequence length after tokenization."
        "The training examples that are longer than this size (input length + output_length) will not be used for training or evaluation."
        "Default to the model max input length for single sentence inputs (take into account special tokens).",
    )
    parser.add_argument(
        '--sort_by_length',
        action='store_true',
        help='Sorts the training set by example length (input length + output_length) to reduce padding and speed up training. Has no effect on accuracy.',
    )
    parser.add_argument("--do_train", action='store_true', help="Whether to run training.")
    parser.add_argument("--do_eval", action='store_true', help="Whether to run eval on the dev set.")
    parser.add_argument(
        "--evaluate_during_training", action='store_true', help="Run evaluation during training at each logging step."
    )
    parser.add_argument("--do_lower_case", action='store_true', help="Set this flag if you are using an uncased model.")

    parser.add_argument("--per_gpu_train_batch_size", default=4, type=int, help="Batch size per GPU/CPU for training.")
    parser.add_argument("--per_gpu_eval_batch_size", default=4, type=int, help="Batch size per GPU/CPU for evaluation.")
    parser.add_argument(
        '--gradient_accumulation_steps',
        type=int,
        default=1,
        help="Number of updates steps to accumulate before performing a backward/update pass.",
    )
    parser.add_argument("--learning_rate", default=5e-5, type=float, help="The initial learning rate for Adam.")
    parser.add_argument("--weight_decay", default=0.0, type=float, help="Weight decay if we apply some.")
    parser.add_argument("--adam_epsilon", default=1e-8, type=float, help="Epsilon for Adam optimizer.")
    parser.add_argument("--max_grad_norm", default=1.0, type=float, help="Max gradient norm.")
    parser.add_argument("--num_train_epochs", default=1, type=int, help="Total number of training epochs to perform.")
    parser.add_argument(
        "--max_steps",
        default=-1,
        type=int,
        help="If > 0: set total number of training steps to perform. Override num_train_epochs.",
    )
    parser.add_argument("--warmup_steps", default=0, type=int, help="Linear warmup over warmup_steps.")
    parser.add_argument(
        "--scheduler",
        default='linear',
        type=str,
        choices=['linear', 'transformer'],
        help="The type of learning rate scheduler to use.",
    )

    parser.add_argument('--logging_steps', type=int, default=50, help="Log every X updates steps.")
    parser.add_argument('--save_steps', type=int, default=50, help="Save checkpoint every X updates steps.")
    parser.add_argument(
        '--save_total_limit',
        type=int,
        default=None,
        help='Limit the total amount of checkpoints, delete the older checkpoints in the output_dir, does not delete by default',
    )
    parser.add_argument(
        "--eval_all_checkpoints",
        action='store_true',
        help="Evaluate all checkpoints starting with the same prefix as model_name_or_path ending and ending with step number",
    )
    parser.add_argument("--no_cuda", action='store_true', help="Avoid using CUDA when available")
    parser.add_argument('--overwrite_output_dir', action='store_true', help="Overwrite the content of the output directory")
    parser.add_argument('--overwrite_cache', action='store_true', help="Overwrite the cached training and evaluation sets")
    parser.add_argument('--seed', type=int, default=42, help="random seed for initialization")

    parser.add_argument(
        '--fp16', action='store_true', help="Whether to use 16-bit (mixed) precision (through NVIDIA apex) instead of 32-bit"
    )
    parser.add_argument(
        '--fp16_opt_level',
        type=str,
        default='O1',
        help="For fp16: Apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']."
        "See details at https://nvidia.github.io/apex/amp.html",
    )
    parser.add_argument("--local_rank", type=int, default=-1, help="For distributed training: local_rank")

    parser.add_argument('--src_lang', type=str, help='source language used for translation task')
    parser.add_argument('--tgt_lang', type=str, help='target language used for translation task')

    parser.add_argument('--debug', action='store_true', help='print intermediate results for debugging')
    parser.add_argument('--no_pretraining', action='store_true', help='Remove all pre-training and train a model from scratch')
    parser.add_argument("--freeze_decoder", action="store_true")
    parser.add_argument("--freeze_encoder", action="store_true")
    parser.add_argument("--freeze_embeds", action="store_true")

    parser.add_argument(
        '--input_column', type=int, default=0, help='The column in the input file which contains the input sentences.'
    )
    parser.add_argument(
        '--gold_column',
        type=int,
        default=1,
        help='The column in the input file which contains the gold sentences. Defaults to --input_column if no gold is available.',
    )

    parser.add_argument(
        '--cache_input_data', action='store_true', help='Cache examples from input data for faster subsequent trainings'
    )

    parser.add_argument(
        '--num_input_chunks',
        default=1,
        type=int,
        help='We split input into multiple chunks, then load and train on each chunk individually',
    )
    parser.add_argument('--delete_after_chunking', action='store_true', help='Delete input file after chunking it')

    parser.add_argument('--no_fast_tokenizer', action='store_true', help='Use slow version of huggingface tokenizer')


def main(args):
    if args.model_type == 'bert' and (
        args.pad_token != '[PAD]' or args.start_special_token != '[SEP]' or args.end_special_token != '[SEP]'
    ):
        raise ValueError("BERT already has its own special tokens [PAD] and [SEP]. You should use them for better results.")
    if args.do_train:
        if args.train_data_file is None:
            raise ValueError(
                "Cannot do training without a training data file. Either supply a file to --train_data_file "
                "or remove the --do_train argument."
            )
        if args.tensorboard_dir is None:
            raise ValueError("Cannot do training without specifying --tensorboard_dir")

    if args.eval_data_file is None and args.do_eval:
        raise ValueError(
            "Cannot do evaluation without an evaluation data file. Either supply a file to --eval_data_file "
            "or remove the --do_eval argument."
        )

    if os.path.exists(args.output_dir) and os.listdir(args.output_dir) and args.do_train:
        # clean all files within the directory
        if args.overwrite_output_dir:
            shutil.rmtree(args.output_dir)
        else:
            raise ValueError(
                "Output directory ({}) already exists and is not empty. Use --overwrite_output_dir to overcome.".format(
                    args.output_dir
                )
            )

    check_args(args)

    if args.gold_column is None:
        args.gold_column = args.input_column

    # Setup CUDA, GPU & distributed training
    if args.local_rank == -1 or args.no_cuda:
        device = torch.device("cuda" if torch.cuda.is_available() and not args.no_cuda else "cpu")
        args.n_gpu = torch.cuda.device_count()
    else:  # Initializes the distributed backend which will take care of sychronizing nodes/GPUs
        torch.cuda.set_device(args.local_rank)
        device = torch.device("cuda", args.local_rank)
        torch.distributed.init_process_group(backend='nccl')
        args.n_gpu = 1
    args.device = device

    # Setup logging
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
        datefmt='%m/%d/%Y %H:%M:%S',
        level=logging.INFO if args.local_rank in [-1, 0] else logging.WARN,
    )
    logger.warning(
        "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
        args.local_rank,
        device,
        args.n_gpu,
        bool(args.local_rank != -1),
        args.fp16,
    )

    # Set seed
    set_seed(args)

    # Load pretrained model and tokenizer
    if args.local_rank not in [-1, 0]:
        torch.distributed.barrier()  # Barrier to make sure only the first process in distributed training download model & vocab

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.config_name if args.config_name else args.model_name_or_path, cache_dir=args.cache_dir if args.cache_dir else None
    )
    tokenizer = tokenizer_class.from_pretrained(
        args.tokenizer_name if args.tokenizer_name else args.model_name_or_path,
        do_lower_case=args.do_lower_case,
        cache_dir=args.cache_dir if args.cache_dir else None,
        use_fast=not args.no_fast_tokenizer,
    )
    if args.block_size <= 0:
        args.block_size = tokenizer.max_len_single_sentence  # Our input block size will be the max possible for the model
    args.block_size = min(args.block_size, tokenizer.max_len_single_sentence)

    if args.no_pretraining:
        # only load model architecture but not the weights
        model = model_class(config)
    else:
        model = model_class.from_pretrained(
            args.model_name_or_path,
            from_tf=bool('.ckpt' in args.model_name_or_path),
            config=config,
            cache_dir=args.cache_dir if args.cache_dir else None,
        )

    add_special_tokens(
        model,
        tokenizer,
        additional_special_tokens=[args.start_special_token, args.end_special_token],
        pad_token=args.pad_token,
    )
    model.to(args.device)

    if args.freeze_embeds:
        freeze_embeds(model)
    if args.freeze_encoder:
        freeze_params(model.get_encoder())
    if args.freeze_decoder:
        if args.model_type in ['bart', 'mbart', 'marian']:
            freeze_params(model.model.decoder)

    if args.local_rank == 0:
        torch.distributed.barrier()  # End of barrier to make sure only the first process in distributed training download model & vocab

    logger.info("Training/evaluation parameters %s", args)

    model_input_prefix = ''
    if args.model_type == 'marian' and args.tgt_lang:
        # TODO check if extra space after pattern is necessary
        model_input_prefix = '>>{}<< '.format(args.tgt_lang)
    elif args.model_type == 't5':
        if args.task == 'translate':
            t5_task = 'translation_{}_to_{}'.format(args.src_lang, args.tgt_lang)
        else:
            t5_task = 'summarization'
        model_input_prefix = config.task_specific_params[t5_task]['prefix']
    args.model_input_prefix = model_input_prefix

    # Training
    if args.do_train:

        if args.num_input_chunks > 1:

            all_input_files = split_file_on_disk(
                args.train_data_file, args.num_input_chunks, delete=args.delete_after_chunking
            )
            global_step, epochs_trained, total_tr_loss = 0, 0, 0.0

            for n in range(args.num_train_epochs):

                for i in range(args.num_input_chunks):

                    if args.local_rank not in [-1, 0]:
                        torch.distributed.barrier()  # Barrier to make sure only the first process in distributed training process the dataset, and the others will use the cache

                    train_dataset = TextDataset(
                        tokenizer, args, file_path=all_input_files[i], block_size=args.block_size, evaluate=True
                    )

                    if args.local_rank == 0:
                        torch.distributed.barrier()

                    global_step, tr_loss = train(
                        args, train_dataset, model, tokenizer, all_input_files[i], True, global_step, epochs_trained
                    )
                    total_tr_loss += tr_loss

                epochs_trained += 1

                logger.info(" global_step = %s, average loss = %s", global_step, total_tr_loss / global_step)

            for file in all_input_files:
                os.remove(file)

        else:
            if args.local_rank not in [-1, 0]:
                torch.distributed.barrier()  # Barrier to make sure only the first process in distributed training process the dataset, and the others will use the cache

            train_dataset = TextDataset(
                tokenizer, args, file_path=args.train_data_file, block_size=args.block_size, evaluate=True
            )

            if args.local_rank == 0:
                torch.distributed.barrier()

            global_step, tr_loss = train(args, train_dataset, model, tokenizer, args.train_data_file, False)

            logger.info(" global_step = %s, average loss = %s", global_step, tr_loss / global_step)

    # Evaluation
    results = {}
    if args.do_eval and args.local_rank in [-1, 0]:
        checkpoints = [args.output_dir]
        if args.eval_all_checkpoints:
            checkpoints = list(
                os.path.dirname(c) for c in sorted(glob.glob(args.output_dir + '/**/' + WEIGHTS_NAME, recursive=True))
            )
            logging.getLogger("transformers.modeling_utils").setLevel(logging.WARN)  # Reduce logging
        logger.info("Evaluate the following checkpoints: %s", checkpoints)
        for checkpoint in checkpoints:
            global_step = checkpoint.split('-')[-1] if len(checkpoints) > 1 else ""
            prefix = checkpoint.split('/')[-1] if checkpoint.find('checkpoint') != -1 else ""

            model = model_class.from_pretrained(checkpoint)
            model.to(args.device)
            result = evaluate(args, model, tokenizer, prefix=prefix)
            result = dict((k + '_{}'.format(global_step), v) for k, v in result.items())
            results.update(result)

    return results



#
# Copyright (c) 2019, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from collections import OrderedDict

from ..data_utils.example import Example
from . import generic_dataset
from .almond_task import BaseAlmondTask
from .base_task import BaseTask
from .generic_dataset import CrossNERDataset
from .registry import register_task


@register_task('multi30k')
class Multi30K(BaseTask):
    @property
    def metrics(self):
        return ['bleu', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        src, trg = ['.' + x for x in self.name.split('.')[1:]]
        return generic_dataset.Multi30k.splits(exts=(src, trg), root=root, **kwargs)


@register_task('iwslt')
class IWSLT(BaseTask):
    @property
    def metrics(self):
        return ['bleu', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        src, trg = ['.' + x for x in self.name.split('.')[1:]]
        return generic_dataset.IWSLT.splits(exts=(src, trg), root=root, **kwargs)


@register_task('squad')
class SQuAD(BaseTask):
    @property
    def metrics(self):
        return ['nf1', 'em', 'nem']

    def get_splits(self, root, **kwargs):
        return generic_dataset.SQuAD.splits(root=root, description=self.name, **kwargs)


@register_task('wikisql')
class WikiSQL(BaseTask):
    @property
    def metrics(self):
        return ['lfem', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        return generic_dataset.WikiSQL.splits(root=root, query_as_question='query_as_question' in self.name, **kwargs)


@register_task('ontonotes')
class OntoNotesNER(BaseTask):
    def get_splits(self, root, **kwargs):
        split_task = self.name.split('.')
        _, _, subtask, nones, counting = split_task
        return generic_dataset.OntoNotesNER.splits(
            subtask=subtask, nones=True if nones == 'nones' else False, root=root, **kwargs
        )


@register_task('woz')
class WoZ(BaseTask):
    @property
    def metrics(self):
        return ['joint_goal_em', 'turn_request_em', 'turn_goal_em', 'avg_dialogue', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        return generic_dataset.WOZ.splits(description=self.name, root=root, **kwargs)


@register_task('multinli')
class MultiNLI(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.MultiNLI.splits(description=self.name, root=root, **kwargs)


@register_task('srl')
class SRL(BaseTask):
    @property
    def metrics(self):
        return ['nf1', 'em', 'nem']

    def get_splits(self, root, **kwargs):
        return generic_dataset.SRL.splits(root=root, **kwargs)


@register_task('snli')
class SNLI(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.SNLI.splits(root=root, **kwargs)


@register_task('schema')
class WinogradSchema(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.WinogradSchema.splits(root=root, **kwargs)


class BaseSummarizationTask(BaseTask):
    @property
    def metrics(self):
        return ['avg_rouge', 'rouge1', 'rouge2', 'rougeL', 'em', 'nem', 'nf1']


@register_task('cnn')
class CNN(BaseSummarizationTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.CNN.splits(root=root, **kwargs)


@register_task('dailymail')
class DailyMail(BaseSummarizationTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.DailyMail.splits(root=root, **kwargs)


@register_task('cnn_dailymail')
class CNNDailyMail(BaseSummarizationTask):
    def get_splits(self, root, **kwargs):
        split_cnn = generic_dataset.CNN.splits(root=root, **kwargs)
        split_dm = generic_dataset.DailyMail.splits(root=root, **kwargs)
        for scnn, sdm in zip(split_cnn, split_dm):
            scnn.examples.extend(sdm)
        return split_cnn


@register_task('sst')
class SST(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.SST.splits(root=root, **kwargs)


@register_task('imdb')
class IMDB(BaseTask):
    def get_splits(self, root, **kwargs):
        kwargs['validation'] = None
        return generic_dataset.IMDb.splits(root=root, **kwargs)


@register_task('zre')
class ZRE(BaseTask):
    @property
    def metrics(self):
        return ['corpus_f1', 'precision', 'recall', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        return generic_dataset.ZeroShotRE.splits(root=root, **kwargs)


@register_task('cross_ner')
class CrossNERTask(BaseAlmondTask):
    politics_labels = [
        'O',
        'B-country',
        'B-politician',
        'I-politician',
        'B-election',
        'I-election',
        'B-person',
        'I-person',
        'B-organisation',
        'I-organisation',
        'B-location',
        'B-misc',
        'I-location',
        'I-country',
        'I-misc',
        'B-politicalparty',
        'I-politicalparty',
        'B-event',
        'I-event',
    ]

    science_labels = [
        'O',
        'B-scientist',
        'I-scientist',
        'B-person',
        'I-person',
        'B-university',
        'I-university',
        'B-organisation',
        'I-organisation',
        'B-country',
        'I-country',
        'B-location',
        'I-location',
        'B-discipline',
        'I-discipline',
        'B-enzyme',
        'I-enzyme',
        'B-protein',
        'I-protein',
        'B-chemicalelement',
        'I-chemicalelement',
        'B-chemicalcompound',
        'I-chemicalcompound',
        'B-astronomicalobject',
        'I-astronomicalobject',
        'B-academicjournal',
        'I-academicjournal',
        'B-event',
        'I-event',
        'B-theory',
        'I-theory',
        'B-award',
        'I-award',
        'B-misc',
        'I-misc',
    ]

    music_labels = [
        'O',
        'B-musicgenre',
        'I-musicgenre',
        'B-song',
        'I-song',
        'B-band',
        'I-band',
        'B-album',
        'I-album',
        'B-musicalartist',
        'I-musicalartist',
        'B-musicalinstrument',
        'I-musicalinstrument',
        'B-award',
        'I-award',
        'B-event',
        'I-event',
        'B-country',
        'I-country',
        'B-location',
        'I-location',
        'B-organisation',
        'I-organisation',
        'B-person',
        'I-person',
        'B-misc',
        'I-misc',
    ]

    literature_labels = [
        'O',
        'B-book',
        'I-book',
        'B-writer',
        'I-writer',
        'B-award',
        'I-award',
        'B-poem',
        'I-poem',
        'B-event',
        'I-event',
        'B-magazine',
        'I-magazine',
        'B-literarygenre',
        'I-literarygenre',
        'B-country',
        'I-country',
        'B-person',
        'I-person',
        'B-location',
        'I-location',
        'B-organisation',
        'I-organisation',
        'B-misc',
        'I-misc',
    ]

    ai_labels = [
        'O',
        'B-field',
        'I-field',
        'B-task',
        'I-task',
        'B-product',
        'I-product',
        'B-algorithm',
        'I-algorithm',
        'B-researcher',
        'I-researcher',
        'B-metrics',
        'I-metrics',
        'B-programlang',
        'I-programlang',
        'B-conference',
        'I-conference',
        'B-university',
        'I-university',
        'B-country',
        'I-country',
        'B-person',
        'I-person',
        'B-organisation',
        'I-organisation',
        'B-location',
        'I-location',
        'B-misc',
        'I-misc',
    ]

    news_labels = ['O', 'B-PER', 'I-PER', 'B-ORG', 'I-ORG', 'B-LOC', 'I-LOC', 'B-MISC', 'I-MISC']

    domain2labels = OrderedDict(
        {
            'politics': politics_labels,
            'science': science_labels,
            'music': music_labels,
            'literature': literature_labels,
            'ai': ai_labels,
            'news': news_labels,
        }
    )

    def __init__(self, name, args):
        self.all_labels = []
        # OrderedDict respect keys order
        for domain, labels in self.domain2labels.items():
            self.all_labels.extend(labels)
        self.label2id = {}
        for label in self.all_labels:
            if label not in self.label2id:
                self.label2id[label] = len(self.label2id)
        self.id2label = {v: k for k, v in self.label2id.items()}
        self.num_labels = len(self.label2id)
        super().__init__(name, args)

    @property
    def metrics(self):
        return ['ner_f1', 'em', 'f1', 'pem']

    def _is_program_field(self, field_name):
        return field_name == 'answer'

    def utterance_field(self):
        return 'context'

    def _make_example(self, parts, dir_name=None, **kwargs):
        example_id = parts[0]
        context = ' '.join(parts[1])
        question = ''
        answer = ' '.join([str(self.label2id[label]) for label in parts[2]])

        return Example.from_raw(
            self.name + '/' + str(example_id), context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def get_splits(self, root, **kwargs):
        return CrossNERDataset.return_splits(name=self.name, path=root, make_example=self._make_example, **kwargs)



# Parts of this file were adopted from https://github.com/huggingface/transformers.
# See the original copyright notice below.

# Copyright 2018 Google AI, Google Brain and Carnegie Mellon University Authors and the HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This script in used for text generation using library models.
It currently supports paraphrasing and translation tasks.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import json
import logging
import math
import os
import re

import numpy as np

# multiprocessing with CUDA
from torch.multiprocessing import Process, set_start_method

from ..data_utils.almond_utils import tokenize_cjk_chars
from ..data_utils.progbar import prange
from ..model_utils.translation import compute_attention, force_replace_quoted_params, replace_quoted_params
from .data_utils import create_features_from_tsv_file, output_heuristics
from .model_utils import compute_metrics

try:
    set_start_method('spawn')
except RuntimeError:
    pass

import torch
from transformers import (
    GPT2_PRETRAINED_CONFIG_ARCHIVE_MAP,
    T5_PRETRAINED_CONFIG_ARCHIVE_MAP,
    BartForConditionalGeneration,
    BartTokenizer,
    GPT2Tokenizer,
    MarianMTModel,
    MarianTokenizer,
    MBart50Tokenizer,
    MBartForConditionalGeneration,
    MT5ForConditionalGeneration,
    PretrainedConfig,
    T5ForConditionalGeneration,
    T5Tokenizer,
)

from ..util import combine_files_on_disk, get_part_path, set_seed, split_file_on_disk
from .data_utils import group_together
from .GPT2Seq2Seq import GPT2Seq2Seq
from .model_utils import check_args
from .transformers_utils import BART_PRETRAINED_CONFIG_ARCHIVE_MAP, MARIAN_PRETRAINED_CONFIG_ARCHIVE_MAP, GenieMBartTokenizer

logging.basicConfig(
    format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s', datefmt='%m/%d/%Y %H:%M:%S', level=logging.INFO
)
logger = logging.getLogger(__name__)

ALL_MODELS = sum(
    (
        tuple(map.keys())
        for map in (
            GPT2_PRETRAINED_CONFIG_ARCHIVE_MAP,
            T5_PRETRAINED_CONFIG_ARCHIVE_MAP,
            BART_PRETRAINED_CONFIG_ARCHIVE_MAP,
            MARIAN_PRETRAINED_CONFIG_ARCHIVE_MAP,
        )
    ),
    (),
)

MODEL_CLASSES = {
    'gpt2': (GPT2Seq2Seq, GPT2Tokenizer, {'bos_token': '<unk>', 'sep_token': '<paraphrase>', 'eos_token': '</paraphrase>'}),
    't5': (T5ForConditionalGeneration, T5Tokenizer, {'bos_token': '<unk>', 'sep_token': '<unk>', 'eos_token': '</s>'}),
    'mt5': (MT5ForConditionalGeneration, T5Tokenizer, {'bos_token': '<unk>', 'sep_token': '<unk>', 'eos_token': '</s>'}),
    'bart': (BartForConditionalGeneration, BartTokenizer, {'bos_token': '<s>', 'sep_token': '<unk>', 'eos_token': '</s>'}),
    'mbart': (
        MBartForConditionalGeneration,
        GenieMBartTokenizer,
        {'bos_token': '<s>', 'sep_token': '<unk>', 'eos_token': '</s>'},
    ),
    'mbart50': (
        MBartForConditionalGeneration,
        MBart50Tokenizer,
        {'bos_token': '<s>', 'sep_token': '<unk>', 'eos_token': '</s>'},
    ),
    'marian': (MarianMTModel, MarianTokenizer, {'bos_token': '<unk>', 'sep_token': '<unk>', 'eos_token': '</s>'}),
}


def parse_argv(parser):
    parser.add_argument(
        "--model_name_or_path",
        default=None,
        type=str,
        required=True,
        help="Path to pre-trained model or shortcut name selected in the list: " + ", ".join(ALL_MODELS),
    )
    parser.add_argument("--input_file", type=str, help="The file from which we read prompts. Defaults to stdin.")
    parser.add_argument(
        '--input_column', type=int, required=True, help='The column in the input file which contains the input sentences.'
    )
    parser.add_argument(
        '--prompt_column',
        type=int,
        default=None,
        help='The column in the input file which contains the text we should start generation from.',
    )
    parser.add_argument(
        '--gold_column',
        type=int,
        default=None,
        help='The column in the input file which contains the gold sentences. Defaults to --input_column if no gold is available.',
    )
    parser.add_argument(
        '--thingtalk_column', type=int, default=None, help='The column in the input file which contains the ThingTalk program.'
    )
    parser.add_argument(
        '--id_column', type=int, default=None, help='The column in the input file which contains the example ID.'
    )
    parser.add_argument(
        "--output_file", type=str, help="When specified, generated text will be written in this file. Defaults to stdout."
    )
    parser.add_argument(
        "--intermediate_file", type=str, default='./paraphrase_tmp.tsv', help="Used to save intermediate results."
    )

    parser.add_argument(
        '--output_prompt',
        action='store_true',
        help='Whether we should include the prompt (specified via --prompt_column or --copy) in the output sequence',
    )
    parser.add_argument("--max_input_length", type=int, default=512, help='Crop longer sentences by this maximum length')
    parser.add_argument(
        "--length", type=int, default=15, help='The generated sentences will have a maximum length of len(input) + arg.length'
    )
    parser.add_argument(
        "--min_output_length",
        type=int,
        default=2,
        help='Will prevent stop tokens from appearing in the first --min_output_length tokens of the generated sentences.',
    )
    parser.add_argument(
        "--skip_heuristics", action='store_true', help='If True, will not replace special word such as NUMBER_0 in the input.'
    )
    parser.add_argument(
        "--is_cased",
        action='store_true',
        help='If True, the trained model is cased, so if --skip_heuristics is not set, we will convert the input to upper case and the output back to lower case.',
    )
    parser.add_argument(
        "--metric_reduction",
        type=str,
        choices=['average', 'max'],
        default='average',
        help="How we should calculate metrics where there are multiple generations per example.",
    )

    parser.add_argument(
        "--shuffle_input",
        action='store_true',
        help='If set, we will shuffle input dataset before processing it'
        'Used mainly with subsampling so we take different portion of data each time',
    )

    parser.add_argument(
        "--pipe_mode", action='store_true', help='If set, we will generate paraphrases of paraphrases of ... as well.'
    )
    # These are generation hyperparameters. Each one can be a list of values in which case, we generate num_samples outputs for each set of hyperparameters.
    parser.add_argument("--num_samples", type=int, nargs='+', default=[1])
    parser.add_argument("--temperature", type=float, nargs='+', default=[1.0], help="temperature of 0 implies greedy sampling")
    parser.add_argument(
        "--repetition_penalty",
        type=float,
        nargs='+',
        default=[1.0],
        help="primarily useful for CTRL model; in that case, use 1.2",
    )
    parser.add_argument("--top_k", type=int, nargs='+', default=[0], help='0 disables top-k filtering')
    parser.add_argument("--top_p", type=float, nargs='+', default=[0.9], help='1.0 disables top-p filtering')
    parser.add_argument("--num_beams", type=int, nargs='+', default=[1], help='1 disables beam seach')
    parser.add_argument(
        "--no_repeat_ngram_size",
        type=int,
        nargs='+',
        default=[0],
        help='ngrams of this size cannot be repeated in the output. 0 disables it.',
    )

    parser.add_argument(
        "--copy",
        type=int,
        default=0,
        help='Number of tokens that will be copied at the beginning of generation. Helps preserve the original meaning of the input sequence.',
    )
    parser.add_argument("--no_cuda", action='store_true', help="Avoid using CUDA when available")
    parser.add_argument('--seed', type=int, default=42, help="random seed for initialization")
    parser.add_argument(
        '--stop_tokens',
        type=str,
        nargs='+',
        default=[],
        help="Tokens (other than the model-specific `eos_token`) at which text generation should be stopped.",
    )
    parser.add_argument('--batch_size', type=int, default=4, help="Batch size for text generation for each GPU.")

    parser.add_argument(
        '--pad_token', type=str, default='<pad>', help='The special token for padding, if tokenizer does not have that'
    )

    parser.add_argument(
        '--cache_dir',
        default='.embeddings',
        type=str,
        help='where to save transforemrs cached models, configs, and tokenizers.',
    )

    parser.add_argument(
        '--trained_model_type', type=str, help='if provided we make sure the loaded model matches the model_type'
    )

    parser.add_argument('--src_lang', type=str, help='source language used for translation task')
    parser.add_argument('--tgt_lang', type=str, help='target language used for translation task')
    parser.add_argument(
        '--output_attentions', action='store_true', help='return self and cross attention weights for seq2seq models'
    )
    parser.add_argument('--output_hidden_states', action='store_true', help='return all hidden states for seq2seq models')

    parser.add_argument(
        '--att_pooling',
        type=str,
        default='max',
        help='pooling used to calculate decoder-encoder attention values across different heads',
    )
    parser.add_argument('--plot_heatmaps', action='store_true', help='whether to plot decoder-encoder attention heatmaps')
    parser.add_argument(
        '--replace_qp', action='store_true', help='replace parameter values after translation with source values'
    )
    parser.add_argument(
        '--force_replace_qp',
        action='store_true',
        help='if we parameters could not be replaced leveraging quotation marks,'
        ' rely purely on attention to find text spans',
    )
    parser.add_argument('--subsample', type=int, default=20000000, help='subsample input datasets')
    parser.add_argument('--task', type=str, required=True, choices=['paraphrase', 'translate'])
    parser.add_argument(
        "--output_example_ids_too", action='store_true', help='Generate two column output with ids in the first column'
    )

    parser.add_argument(
        '--mask_tokens', action='store_true', help='mask input tokens and infill them using denoising pretrained model'
    )
    parser.add_argument(
        '--mask_token_prob', type=float, default=0.15, help='Probability of an input token being masked in the sentence'
    )
    parser.add_argument(
        '--delete_tokens',
        action='store_true',
        help='delete input tokens and infill them using denoising pretrained model'
        'In contrast to token masking, the model should decide which positions have missing inputs',
    )
    parser.add_argument(
        '--delete_token_prob', type=float, default=0.15, help='Probability of an input token being deleted in the sentence'
    )
    parser.add_argument(
        '--infill_text', action='store_true', help='mask consecutive tokens and infill them using denoising pretrained model'
    )
    parser.add_argument(
        '--num_text_spans', type=int, default=3, help='number of text spans to sample for text infilling method'
    )
    parser.add_argument('--infill_max_tries', type=int, default=3, help='Maximum number of tries to find an appropriate span')

    parser.add_argument(
        '--permute_sentences',
        action='store_true',
        help='divide document into sentences based on fill stops and'
        'permutate them. Use this only if input has multiple sentences.',
    )
    parser.add_argument(
        '--rotate_sentence',
        action='store_true',
        help='a pivot token is chosen randomly, and sentence is rotated so new sentence start with pivot token',
    )

    parser.add_argument(
        '--fp16',
        action='store_true',
        help="Whether to use 16-bit (mixed) precision (through NVIDIA apex) instead of 32-bit. On certain GPUs (e.g. Nvidia V100) improves the inference speed",
    )
    parser.add_argument(
        '--fp16_opt_level',
        type=str,
        default='O1',
        help="For fp16: Apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']."
        "See details at https://nvidia.github.io/apex/amp.html",
    )

    parser.add_argument('--verbose', action='store_true', help='log additional information for debugging purposes')

    parser.add_argument('--no_fast_tokenizer', action='store_true', help='Use slow version of huggingface tokenizer')


def main(args):
    hyperparameters = [
        'num_samples',
        'temperature',
        'top_k',
        'top_p',
        'repetition_penalty',
        'num_beams',
        'no_repeat_ngram_size',
    ]
    max_hyperparameter_len = max([len(getattr(args, h)) for h in hyperparameters])
    valid_len = [1, max_hyperparameter_len]
    for h in hyperparameters:
        if len(getattr(args, h)) not in valid_len:
            logger.error('Hyperparameters should either have the same number of values as others or have exactly one value.')
        # If only one value is provided, use the same value for all samples
        setattr(args, h, getattr(args, h) * (max_hyperparameter_len // len(getattr(args, h))))

    logger.info(
        'Will output %d sequences for each input.', sum(args.num_samples) if not args.pipe_mode else np.prod(args.num_samples)
    )
    logger.info('Effective batch size for each device is %d', args.batch_size * max(args.num_samples))

    # TODO using intermediate files for pipe_mode is not clean. It needs to change.
    if args.pipe_mode:
        intermediate_files = [args.input_file] + [args.intermediate_file + str(i) for i in range(max_hyperparameter_len)]
        for i in range(max_hyperparameter_len):
            copy_args = copy.copy(args)
            for h in hyperparameters:
                setattr(copy_args, h, [getattr(args, h)[i]])
            copy_args.input_file = intermediate_files[i]
            copy_args.output_file = intermediate_files[i + 1]
            run_multi_process_generation(copy_args)
        all_outputs = group_together(intermediate_files[1:], args.num_samples)
        for file_path in intermediate_files[1:]:
            os.remove(file_path)
        if args.output_file is not None:
            if not os.path.exists(os.path.dirname(args.output_file)):
                os.makedirs(os.path.dirname(args.output_file), exist_ok=False)
            with open(args.output_file, 'w') as output_file:
                for output in all_outputs:
                    for text in output:
                        output_file.write(text + '\n')
        else:
            print(json.dumps(all_outputs, indent=2))
    else:
        run_multi_process_generation(args)


def run_multi_process_generation(args):
    config = PretrainedConfig.from_pretrained(args.model_name_or_path, cache_dir=args.cache_dir)

    # get model type from saved config
    if hasattr(config, 'model_type'):
        args.model_type = getattr(config, 'model_type')
        if args.model_type == 'mbart' and '-50-' in args.model_name_or_path:
            args.model_type = 'mbart50'
    else:
        raise ValueError('Model should be either GPT2, BART, MBART, or Marian')

    # check arguments validity
    check_args(args)

    if sum([args.mask_tokens, args.delete_tokens, args.infill_text, args.permute_sentences, args.rotate_sentence]) >= 2:
        raise ValueError('Mixing denoising techniques is unlikely to work. Please use one method per run')

    if (args.mask_tokens or args.delete_tokens or args.rotate_sentence) and args.model_type == 'mbart':
        raise ValueError(
            'MBART is pretrained only with text_infilling and permute_sentences noising methods. '
            'Applying other noising techniques is unlikely to work'
        )

    if args.trained_model_type and args.trained_model_type != '' and args.model_type != args.trained_model_type:
        raise ValueError('The loaded model type does not match with what the user provided')

    if args.prompt_column is not None and args.copy is not None and args.copy != 0:
        raise ValueError(
            'Cannot copy from the input and use prompt at the same time. Disable either --copy or --prompt_column.'
        )

    if args.gold_column is None:
        args.gold_column = args.input_column
    args.device = torch.device("cuda" if torch.cuda.is_available() and not args.no_cuda else "cpu")
    args.n_gpu = torch.cuda.device_count()

    if args.output_file is not None:
        if not os.path.exists(os.path.dirname(args.output_file)):
            os.makedirs(os.path.dirname(args.output_file), exist_ok=False)

    set_seed(args)

    if args.n_gpu > 1:
        if args.input_file is None:
            raise ValueError('Cannot use multiple GPUs when reading from stdin. You should provide an --input_file')
        logger.info('Running generation in parallel on {} GPUs'.format(args.n_gpu))
        # Independent multi-GPU generation
        all_processes = []
        all_input_files = split_file_on_disk(args.input_file, args.n_gpu)
        for gpu_idx in range(args.n_gpu):
            copy_args = copy.copy(args)
            if torch.cuda.is_available() and not args.no_cuda:
                copy_args.device = torch.device("cuda:" + str(gpu_idx))
            copy_args.n_gpu = 1
            copy_args.input_file = all_input_files[gpu_idx]
            copy_args.output_file = get_part_path(args.output_file, gpu_idx)

            p = Process(target=run_single_process_generation, args=(copy_args, config))
            all_processes.append(p)
            p.start()

        for p in all_processes:
            p.join()

        for file in all_input_files:
            os.remove(file)
        combine_files_on_disk(args.output_file, args.n_gpu, line_group_size=sum(args.num_samples), delete=True)

    else:
        run_single_process_generation(args, config)


def run_single_process_generation(args, config):
    model_class, tokenizer_class, special_tokens = MODEL_CLASSES[args.model_type]

    output_attentions = args.output_attentions
    output_hidden_states = args.output_hidden_states

    model = model_class.from_pretrained(
        args.model_name_or_path,
        output_attentions=output_attentions,
        output_hidden_states=output_hidden_states,
        cache_dir=args.cache_dir,
    )
    model.to(args.device)

    if args.fp16:
        try:
            from apex import amp
        except ImportError:
            raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use fp16 training.")
        model = amp.initialize(model, opt_level=args.fp16_opt_level)

    model.eval()
    tokenizer = tokenizer_class.from_pretrained(
        args.model_name_or_path, cache_dir=args.cache_dir, use_fast=not args.no_fast_tokenizer
    )
    eos_token_id = tokenizer.convert_tokens_to_ids(special_tokens['eos_token'])
    sep_token_id = tokenizer.convert_tokens_to_ids(special_tokens['sep_token'])

    if tokenizer.pad_token is None:
        # this assigns pad token but doesn't add it to the vocabulary
        tokenizer.pad_token = args.pad_token

    pad_token_id = tokenizer.convert_tokens_to_ids(tokenizer.pad_token)

    if pad_token_id is None:
        logger.error('Your tokenizer does not have a padding token')

    if args.model_type == 'gpt2':
        model.set_token_ids(eos_token_id=eos_token_id, sep_token_id=sep_token_id, pad_token_id=pad_token_id)

    logger.info(args)

    model_input_prefix = ''
    if args.model_type == 'marian' and args.tgt_lang:
        # TODO check if extra space after pattern is necessary
        model_input_prefix = '>>{}<< '.format(args.tgt_lang)
    elif args.model_type == 't5':
        if args.task == 'translate':
            t5_task = 'translation_{}_to_{}'.format(args.src_lang, args.tgt_lang)
        else:
            t5_task = 'summarization'
        model_input_prefix = config.task_specific_params[t5_task]['prefix']

    masking_token = getattr(tokenizer, 'mask_token', '<mask>')

    (
        all_input_sequences,
        all_input_sequence_lengths,
        all_example_ids,
        all_context_ids,
        estimated_output_lengths,
        all_golds,
        reverse_maps,
        all_prompt_ids,
    ) = create_features_from_tsv_file(
        file_path=args.input_file,
        tokenizer=tokenizer,
        input_column=args.input_column,
        gold_column=args.gold_column,
        id_column=args.id_column,
        prompt_column=args.prompt_column,
        thingtalk_column=args.thingtalk_column,
        copy=args.copy,
        sep_token_id=sep_token_id,
        skip_heuristics=args.skip_heuristics,
        is_cased=args.is_cased,
        model_type=args.model_type,
        src_lang=args.src_lang,
        subsample=args.subsample,
        shuffle_input=args.shuffle_input,
        task=args.task,
        model_input_prefix=model_input_prefix,
        max_input_length=args.max_input_length,
        mask_tokens=args.mask_tokens,
        mask_token_prob=args.mask_token_prob,
        masking_token=masking_token,
        infill_max_tries=args.infill_max_tries,
        delete_tokens=args.delete_tokens,
        delete_token_prob=args.delete_token_prob,
        infill_text=args.infill_text,
        num_text_spans=args.num_text_spans,
        permute_sentences=args.permute_sentences,
        rotate_sentence=args.rotate_sentence,
    )

    # sort contexts based on their context length so that less generated tokens are thrown away and generation can be done faster
    (
        estimated_output_lengths,
        all_input_sequence_lengths,
        all_input_sequences,
        all_context_ids,
        original_order,
        reverse_maps,
        all_prompt_ids,
    ) = tuple(
        zip(
            *sorted(
                list(
                    zip(
                        estimated_output_lengths,
                        all_input_sequence_lengths,
                        all_input_sequences,
                        all_context_ids,
                        range(len(all_context_ids)),
                        reverse_maps,
                        all_prompt_ids,
                    )
                ),
                reverse=True,
            )
        )
    )
    all_outputs = []

    stop_token_ids = [tokenizer.convert_tokens_to_ids(stop_token) for stop_token in args.stop_tokens]

    batch_idx = 0
    for batch in prange(math.ceil(len(all_context_ids) / args.batch_size)):
        batch_slice = (batch * args.batch_size, min((batch + 1) * args.batch_size, len(all_context_ids)))
        batch_size = batch_slice[1] - batch_slice[0]
        batch_context_tokens = all_context_ids[batch_slice[0] : batch_slice[1]]
        batch_reverse_maps = reverse_maps[batch_slice[0] : batch_slice[1]]
        batch_prompt_tokens = all_prompt_ids[batch_slice[0] : batch_slice[1]]

        if args.model_type == 'gpt2':
            batch_context_tensor = torch.tensor(
                model.pad_to_max_length(batch_context_tokens), dtype=torch.long, device=args.device
            )
            attention_mask = None
        else:
            padded_batch_context_tokens = []
            max_length = max([len(s) for s in batch_context_tokens])
            for i in range(len(batch_context_tokens)):
                padded_batch_context_tokens.append(
                    batch_context_tokens[i] + [pad_token_id] * (max_length - len(batch_context_tokens[i]))
                )
            batch_context_tensor = torch.tensor(padded_batch_context_tokens, dtype=torch.long, device=args.device)
            attention_mask = (batch_context_tensor != pad_token_id).to(torch.long)

        if args.model_type == 'mbart':
            decoder_start_token_id = tokenizer.lang_code_to_id[args.tgt_lang]
            model.config.decoder_start_token_id = decoder_start_token_id
        else:
            decoder_start_token_id = None

        if args.model_type == 'mbart50':
            forced_bos_token_id = tokenizer.lang_code_to_id[args.tgt_lang]
        else:
            forced_bos_token_id = None

        max_length = batch_context_tensor.shape[1] + args.length

        batch_outputs = [[] for _ in range(batch_size)]
        for hyperparameter_idx in range(len(args.temperature)):
            outputs = model.generate(
                input_ids=batch_context_tensor,
                bad_words_ids=None,
                attention_mask=attention_mask,
                decoder_start_token_id=decoder_start_token_id,
                forced_bos_token_id=forced_bos_token_id,
                min_length=args.min_output_length,
                max_length=max_length,
                num_beams=args.num_beams[hyperparameter_idx],
                top_k=args.top_k[hyperparameter_idx],
                top_p=args.top_p[hyperparameter_idx],
                early_stopping=True,
                num_return_sequences=args.num_samples[hyperparameter_idx],
                repetition_penalty=args.repetition_penalty[hyperparameter_idx],
                no_repeat_ngram_size=args.no_repeat_ngram_size[hyperparameter_idx],
                do_sample=args.temperature[hyperparameter_idx] != 0,
                temperature=args.temperature[hyperparameter_idx]
                if args.temperature[hyperparameter_idx] > 0
                else 1.0,  # if temperature==0, we do not sample
                eos_token_id=eos_token_id,
                pad_token_id=pad_token_id,
                use_cache=True,
                output_attentions=output_attentions,
                return_dict_in_generate=True,
            )

            decoded = outputs.sequences
            cross_attentions = getattr(outputs, 'cross_attentions', None)

            if cross_attentions is not None:
                # stack tensors to shape (max_output_length, num_layers, batch_size, num_heads, 1, max_input_length)
                cross_attentions = torch.stack(([torch.stack(tuple) for tuple in cross_attentions]))

                # reshape to (num_layers, batch_size, num_heads, max_output_length, max_input_length)
                cross_attentions = cross_attentions.squeeze(4)
                cross_attentions = cross_attentions.permute(1, 2, 3, 0, 4).contiguous()

            if not isinstance(decoded, list):
                decoded = decoded[:, :].tolist()
            for i, out in enumerate(decoded):
                if 'bart' in args.model_type:
                    out = out[1:]  # remove </s> token at the beginning for bart, mbart, mbart50
                sample_index = (i // args.num_samples[hyperparameter_idx]) % batch_size
                if not args.output_prompt:
                    out = out[len(batch_prompt_tokens[sample_index]) :]
                min_index = len(out) - 1
                for stop_token_id in stop_token_ids + [eos_token_id]:
                    try:
                        index = out.index(stop_token_id)
                        min_index = min(index, min_index)
                    except ValueError:
                        pass

                ### include eos_token too; it will get removed during decoding
                min_index = min_index + 1
                out_cropped = out[:min_index]

                if args.task == 'translate' and cross_attentions is not None:
                    src_tokens = tokenizer.convert_ids_to_tokens(batch_context_tensor[sample_index])
                    tgt_tokens = tokenizer.convert_ids_to_tokens(out_cropped)

                    # get last layer attention vectors
                    # TODO: get penultimate layer of attention vectors
                    layer_attention = cross_attentions[-1, ...]
                    sample_layer_attention = layer_attention[sample_index, :, :, :]

                    if (
                        tgt_tokens[0] in [tokenizer.pad_token, special_tokens['bos_token'], special_tokens['sep_token']]
                        or (decoder_start_token_id and tgt_tokens[0] == tokenizer.id_to_lang_code[decoder_start_token_id])
                        or (forced_bos_token_id and tgt_tokens[0] == tokenizer.id_to_lang_code[forced_bos_token_id])
                    ):
                        # shift target tokens left to match the attention positions
                        tgt_tokens = tgt_tokens[1:]

                    while src_tokens[-1] == tokenizer.pad_token:
                        # remove all padding from src
                        src_tokens = src_tokens[:-1]
                    if src_tokens[-1] == special_tokens['sep_token']:
                        # remove trailing sep token
                        src_tokens = src_tokens[:-1]
                    if src_tokens[-1] == special_tokens['eos_token']:
                        # remove end token for better heatmap representation
                        src_tokens = src_tokens[:-1]

                    # remove language code from the beginning of src_tokens and shift layer_attention
                    len_prefix_wp = len(tokenizer.tokenize(model_input_prefix))
                    src_tokens = src_tokens[len_prefix_wp:]
                    sample_layer_attention = sample_layer_attention[:, :, len_prefix_wp:]

                    # crop to match src and tgt new lengths
                    sample_layer_attention = sample_layer_attention[:, : len(tgt_tokens), : len(src_tokens)]

                    sample_layer_attention_pooled = compute_attention(sample_layer_attention, args.att_pooling)

                    if args.plot_heatmaps:
                        import matplotlib.pyplot as plt
                        import seaborn as sns

                        src_tokens = [token.lower() for token in src_tokens]
                        tgt_tokens = [token.lower() for token in tgt_tokens]
                        g = sns.heatmap(
                            torch.log(sample_layer_attention_pooled), xticklabels=src_tokens, yticklabels=tgt_tokens
                        )
                        g.set_xticklabels(g.get_xmajorticklabels(), fontsize=12)
                        g.set_yticklabels(g.get_ymajorticklabels(), fontsize=12)
                        if args.output_file is not None:
                            plt.savefig(
                                os.path.join(
                                    os.path.dirname(args.output_file), 'heatmap_{}'.format(batch_idx * batch_size + i)
                                )
                            )
                        plt.show()

                    # remove end token if present
                    if tgt_tokens[-1] in [special_tokens['bos_token'], special_tokens['eos_token']]:
                        tgt_tokens = tgt_tokens[:-1]

                    if args.replace_qp:
                        text, is_replaced = replace_quoted_params(
                            src_tokens, tgt_tokens, tokenizer, sample_layer_attention_pooled
                        )
                        if not is_replaced and args.force_replace_qp:
                            text = force_replace_quoted_params(
                                src_tokens, tgt_tokens, tokenizer, sample_layer_attention_pooled
                            )
                    else:
                        text = tokenizer.convert_tokens_to_string(tgt_tokens)
                else:
                    text = tokenizer.decode(out_cropped, clean_up_tokenization_spaces=False, skip_special_tokens=True)

                text = re.sub('\s\s+', ' ', text)  # remove duplicate white spaces
                text = text.strip()

                text = tokenize_cjk_chars(text)

                if not args.skip_heuristics:
                    text = output_heuristics(text, batch_reverse_maps[sample_index])
                batch_outputs[sample_index].append(text)

        all_outputs.extend(batch_outputs)
        if batch_idx == 0 and args.verbose:
            logger.info('First batch output: %s', str(all_outputs))
        batch_idx += 1

    # sort the results back to their original order
    _, all_outputs = tuple(zip(*sorted(list(zip(original_order, all_outputs)))))

    if args.output_file is not None:
        with open(args.output_file, 'w') as output_file:
            for i, output in enumerate(all_outputs):
                for j, text in enumerate(output):
                    # if num_samples is 1 keep the original id
                    if len(output) == 1:
                        id_ = all_example_ids[i]
                    else:
                        id_ = '{}-{}'.format(all_example_ids[i], j)
                    if args.output_example_ids_too:
                        output_file.write('\t'.join([id_, text]) + '\n')
                    else:
                        output_file.write(text + '\n')

    else:
        print(json.dumps(all_outputs, indent=2))

    metrics = compute_metrics(all_outputs, all_golds, reduction=args.metric_reduction)
    logger.info('Average BLEU score = %.2f', metrics['bleu'])
    logger.info('Exact match score = %.2f', metrics['em'])



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import logging
import os
import random
import re
import shutil
import time
from json.decoder import JSONDecodeError
from typing import List, Optional

import numpy as np
import torch
import ujson
from torch.functional import Tensor
from transformers import MarianConfig, MBartConfig
from transformers.models.mbart.tokenization_mbart50 import FAIRSEQ_LANGUAGE_CODES

from .data_utils.almond_utils import token_type_regex
from .data_utils.example import NumericalizedExamples
from .data_utils.iterator import LengthSortedIterator
from .paraphrase.transformers_utils import MARIAN_GROUP_MEMBERS

logger = logging.getLogger(__name__)

ENTITY_MATCH_REGEX = re.compile('^([A-Z].*)_[0-9]+$')


class SpecialTokenMap:
    def __init__(self, pattern, forward_func, backward_func=None):
        """
        Inputs:
            pattern: a regex pattern
            forward_func: a function with signature forward_func(str) -> str
            backward_func: a function with signature backward_func(str) -> list[str]
        """
        if isinstance(forward_func, list):
            self.forward_func = lambda x: forward_func[int(x) % len(forward_func)]
        else:
            self.forward_func = forward_func

        if isinstance(backward_func, list):
            self.backward_func = lambda x: backward_func[int(x) % len(backward_func)]
        else:
            self.backward_func = backward_func

        self.pattern = pattern

    def forward(self, s: str):
        reverse_map = []
        matches = re.finditer(self.pattern, s)
        if matches is None:
            return s, reverse_map
        for match in matches:
            occurrence = match.group(0)
            parameter = match.group(1)
            replacement = self.forward_func(parameter)
            s = s.replace(occurrence, replacement)
            reverse_map.append((self, occurrence))
        return s, reverse_map

    def backward(self, s: str, occurrence: str):
        match = re.match(self.pattern, occurrence)
        parameter = match.group(1)
        if self.backward_func is None:
            list_of_strings_to_match = [self.forward_func(parameter)]
        else:
            list_of_strings_to_match = sorted(self.backward_func(parameter), key=lambda x: len(x), reverse=True)
        for string_to_match in list_of_strings_to_match:
            l_ = [' ' + string_to_match + ' ', string_to_match + ' ', ' ' + string_to_match]
            o_ = [' ' + occurrence + ' ', occurrence + ' ', ' ' + occurrence]
            new_s = s
            for i in range(len(l_)):
                new_s = re.sub(l_[i], o_[i], s, flags=re.IGNORECASE)
                if s != new_s:
                    break
            if s != new_s:
                s = new_s
                break

        return s


class ConfidenceFeatures:
    """
    Contains all necessary features that are useful for calculating confidence of a single generated output
    """

    def __init__(
        self,
        drop_logits: List[Tensor],
        drop_probs: List[Tensor],
        gold_answer: Tensor,
        prediction: Tensor,
        nodrop_logits: Tensor,
        nodrop_probs: Tensor,
        nodrop_entropies: Tensor,
        context: Tensor,
        nodrop_top1_probs: Tensor = None,
        nodrop_top2_probs: Tensor = None,
        drop_top1_probs: List[Tensor] = None,
        drop_top2_probs: List[Tensor] = None,
    ):
        """
        Inputs:
            droplogits: logits after MC dropout
            gold_answer: includes BOS and EOS tokens, but no PAD tokens
            prediction: includes BOS and EOS tokens, but no PAD tokens
            nodrop_logits: logits for this prediction that are obtained WITHOUT activating model's dropout
            nodrop_top1_probs, nodrop_top2_probs: highest and second highest probabilities of the next token, given that the previous token was from `prediction`
        """

        # store the results of MC dropout if provided
        self.drop_logits = drop_logits
        self.drop_probs = drop_probs
        self.drop_top1_probs = drop_top1_probs
        self.drop_top2_probs = drop_top2_probs

        if drop_logits is not None:
            self.drop_logits = torch.stack(drop_logits, dim=0).cpu()
        if drop_probs is not None:
            self.drop_probs = torch.stack(drop_probs, dim=0).cpu()
        if drop_top1_probs is not None:
            self.drop_top1_probs = torch.stack(drop_top1_probs, dim=0).cpu()
        if drop_top2_probs is not None:
            self.drop_top2_probs = torch.stack(drop_top2_probs, dim=0).cpu()

        # store the results of non-dropout forward pass, if provided
        self.nodrop_logits = nodrop_logits
        self.nodrop_probs = nodrop_probs
        self.nodrop_entropies = nodrop_entropies
        self.nodrop_top1_probs = nodrop_top1_probs
        self.nodrop_top2_probs = nodrop_top2_probs

        if nodrop_logits is not None:
            self.nodrop_logits = nodrop_logits.cpu()
        if nodrop_probs is not None:
            self.nodrop_probs = nodrop_probs.cpu()
        if nodrop_entropies is not None:
            self.nodrop_entropies = nodrop_entropies.cpu()
        if nodrop_top1_probs is not None:
            self.nodrop_top1_probs = nodrop_top1_probs.cpu()
        if nodrop_top2_probs is not None:
            self.nodrop_top2_probs = nodrop_top2_probs.cpu()

        self.prediction = prediction
        self.gold_answer = gold_answer
        self.first_mistake = ConfidenceFeatures.find_first_mistake(gold_answer, prediction)
        self.label = self.first_mistake == -1
        self.context = context

    @property
    def mc_dropout_num(self):
        if self.drop_logits is None:
            return 0
        else:
            return self.drop_logits.shape[0]

    @staticmethod
    def find_first_mistake(gold_answer: Tensor, prediction: Tensor):
        """
        Inputs:
            gold_answer: includes BOS and EOS tokens, but no PAD tokens
            prediction: includes BOS and EOS tokens, but no PAD tokens
        Returns:
            The index of the first token where `gold_answer` and `prediction` differ, or -1 if they are equal. Ignores BOS, so the minimum possible value is 0.
        """
        # print('gold_answer = ', gold_answer)
        # print('prediction = ', prediction)
        # skip BOS
        gold_answer = gold_answer[1:]
        prediction = prediction[1:]

        for i in range(min(len(gold_answer), len(prediction))):
            if gold_answer[i] != prediction[i]:
                return i
        if len(gold_answer) != len(prediction):
            # one is a strict prefix of the other
            return min(len(gold_answer), len(prediction))
        return -1

    def __repr__(self) -> str:
        return (
            '<Confidence: drop_logits='
            + str(self.drop_logits)
            + ', drop_probs='
            + str(self.drop_probs)
            + ', first_mistake='
            + str(self.first_mistake)
            + ', nodrop_logits='
            + str(self.nodrop_logits)
            + ', nodrop_probs='
            + str(self.nodrop_probs)
            + ', nodrop_entropies='
            + str(self.nodrop_entropies)
            + ', context='
            + str(self.context)
            + ', label='
            + str(self.label)
            + '>'
        )


class GenerationOutput:
    """
    Contains all the information that the generation function may need to output
    """

    def __init__(
        self,
        loss: Optional[float] = None,
        example_ids: Optional[List] = None,
        predictions: Optional[List] = None,
        answers: Optional[List] = None,
        contexts: Optional[List] = None,
        confidence_features: Optional[List] = None,
        confidence_scores: Optional[List] = None,
    ):
        self.loss = loss
        self.example_ids = example_ids
        self.predictions = predictions
        self.answers = answers
        self.contexts = contexts
        self.confidence_features = confidence_features
        self.confidence_scores = confidence_scores


def remove_thingtalk_quotes(thingtalk):
    quote_values = []
    while True:
        # print('before: ', thingtalk)
        l1 = thingtalk.find('"')
        if l1 < 0:
            break
        l2 = thingtalk.find('"', l1 + 1)
        if l2 < 0:
            # ThingTalk code is not syntactic
            return thingtalk, None
        quote_values.append(thingtalk[l1 + 1 : l2].strip())
        thingtalk = thingtalk[:l1] + '<temp>' + thingtalk[l2 + 1 :]
        # print('after: ', thingtalk)
    thingtalk = thingtalk.replace('<temp>', '""')
    return thingtalk, quote_values


def find_span_type(program, begin_index, end_index):

    if begin_index > 1 and program[begin_index - 2] == 'location:':
        span_type = 'LOCATION'
    elif end_index == len(program) - 1 or not program[end_index + 1].startswith('^^'):
        span_type = 'QUOTED_STRING'
    else:
        if program[end_index + 1] == '^^tt:hashtag':
            span_type = 'HASHTAG'
        elif program[end_index + 1] == '^^tt:username':
            span_type = 'USERNAME'
        else:
            span_type = 'GENERIC_ENTITY_' + program[end_index + 1][2:]

        end_index += 1

    return span_type, end_index


def requote_program(program):

    program = program.split(' ')
    requoted = []

    in_string = False
    begin_index = 0
    i = 0
    while i < len(program):
        token = program[i]
        if token == '"':
            in_string = not in_string
            if in_string:
                begin_index = i + 1
            else:
                span_type, end_index = find_span_type(program, begin_index, i)
                requoted.append(span_type)
                i = end_index

        elif not in_string:
            entity_match = ENTITY_MATCH_REGEX.match(token)
            if entity_match is not None:
                requoted.append(entity_match[1])
            elif token != 'location:':
                requoted.append(token)

        i += 1

    return ' '.join(requoted)


def tokenizer(s):
    return s.split()


def mask_special_tokens(string: str):
    exceptions = [match.group(0) for match in re.finditer('[A-Za-z:_.]+_[0-9]+', string)]
    for e in exceptions:
        string = string.replace(e, '<temp>', 1)
    return string, exceptions


def unmask_special_tokens(string: str, exceptions: list):
    for e in exceptions:
        string = string.replace('<temp>', e, 1)
    return string


def detokenize(string: str):
    string, exceptions = mask_special_tokens(string)
    tokens = ["'d", "n't", "'ve", "'m", "'re", "'ll", ".", ",", "?", "!", "'s", ")", ":", "-"]
    for t in tokens:
        string = string.replace(' ' + t, t)
    string = string.replace("( ", "(")
    string = string.replace('gon na', 'gonna')
    string = string.replace('wan na', 'wanna')
    string = unmask_special_tokens(string, exceptions)
    return string


def tokenize(string: str):
    string, exceptions = mask_special_tokens(string)
    tokens = ["'d", "n't", "'ve", "'m", "'re", "'ll", ".", ",", "?", "!", "'s", ")", ":"]
    for t in tokens:
        string = string.replace(t, ' ' + t)
    string = string.replace("(", "( ")
    string = string.replace('gonna', 'gon na')
    string = string.replace('wanna', 'wan na')
    string = unmask_special_tokens(string, exceptions)
    string = re.sub('([A-Za-z:_.]+_[0-9]+)-', r'\1 - ', string)  # add space before and after hyphen, e.g. "NUMBER_0-hour"
    string = re.sub('\s+', ' ', string)  # remove duplicate spaces
    return string.strip()


def lower_case(string):
    string, exceptions = mask_special_tokens(string)
    string = string.lower()
    string = unmask_special_tokens(string, exceptions)
    return string


def get_number_of_lines(file_path):
    count = 0
    with open(file_path) as f:
        for line in f:
            count += 1
    return count


def get_part_path(path, part_idx):
    if path.endswith(os.path.sep):
        has_separator = True
        path = path[:-1]
    else:
        has_separator = False
    return path + '_part' + str(part_idx + 1) + (os.path.sep if has_separator else '')


def split_folder_on_disk(folder_path, num_splits):
    new_folder_paths = [get_part_path(folder_path, part_idx) for part_idx in range(num_splits)]
    for subdir, dirs, files in os.walk(folder_path):
        for file in files:
            new_file_paths = [
                os.path.join(subdir.replace(folder_path, new_folder_paths[part_idx]), file) for part_idx in range(num_splits)
            ]
            split_file_on_disk(os.path.join(subdir, file), num_splits, output_paths=new_file_paths)
    return new_folder_paths


def split_file_on_disk(file_path, num_splits, output_paths=None, delete=False):
    """ """

    all_output_paths = []
    all_output_files = []
    for part_idx in range(num_splits):
        if output_paths is None:
            output_path = get_part_path(file_path, part_idx)
        else:
            output_path = output_paths[part_idx]
        all_output_paths.append(output_path)
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        all_output_files.append(open(output_path, 'w'))

    with open(file_path, 'r') as input_file:
        output_file_idx = 0
        for line in input_file:
            all_output_files[output_file_idx].write(line)
            output_file_idx = (output_file_idx + 1) % len(all_output_files)

    for f in all_output_files:
        f.close()

    if delete:
        os.remove(file_path)

    return all_output_paths


def combine_folders_on_disk(folder_path_prefix, num_files, line_group_size, delete=False):
    folder_paths = [get_part_path(folder_path_prefix, part_idx) for part_idx in range(num_files)]
    new_to_olds_map = {}
    for i in range(num_files):
        for subdir, dirs, files in os.walk(folder_paths[i]):
            for file in files:
                new_file_path = os.path.join(subdir.replace(folder_paths[i], folder_path_prefix), file)
                if new_file_path not in new_to_olds_map:
                    new_to_olds_map[new_file_path] = []
                new_to_olds_map[new_file_path].append(os.path.join(subdir, file))

    for new, olds in new_to_olds_map.items():
        os.makedirs(os.path.dirname(new), exist_ok=True)
        with open(new, 'w') as combined_file:
            if new.endswith('.json'):
                new_json = None
                for old in olds:
                    with open(old, 'r') as f:
                        if new_json is None:
                            try:
                                new_json = json.load(f)
                            except JSONDecodeError:
                                f.seek(0)
                                logger.info('Failed to read json file %s with content:\n %s', old, f.read())
                        else:
                            for k, v in json.load(f).items():
                                new_json[k] += v
                for k, v in new_json.items():
                    new_json[k] /= float(num_files)
                json.dump(new_json, combined_file)
            else:
                all_old_file_contents = []
                for old in olds:
                    with open(old, 'r') as f:
                        all_old_file_contents.append([line for line in f])
                old_file_idx = 0
                all_indices = [0] * len(all_old_file_contents)
                finished_reading = [False] * len(all_old_file_contents)
                while True:
                    if finished_reading[old_file_idx]:
                        old_file_idx = (old_file_idx + 1) % len(all_old_file_contents)
                        continue
                    for i in range(line_group_size):
                        line = all_old_file_contents[old_file_idx][all_indices[old_file_idx]]
                        combined_file.write(line)
                        all_indices[old_file_idx] += 1
                    if all_indices[old_file_idx] == len(all_old_file_contents[old_file_idx]):
                        finished_reading[old_file_idx] = True
                        if all(finished_reading):
                            break
                    old_file_idx = (old_file_idx + 1) % len(all_old_file_contents)

    if delete:
        for folder in folder_paths:
            shutil.rmtree(folder)


def combine_files_on_disk(file_path_prefix, num_files, line_group_size, delete=False):
    all_input_file_contents = []
    all_input_file_paths = []
    for i in range(num_files):
        input_file_path = get_part_path(file_path_prefix, i)
        all_input_file_paths.append(input_file_path)
        with open(input_file_path, 'r') as f:
            all_input_file_contents.append([line for line in f])

    all_indices = [0] * len(all_input_file_contents)
    finished_reading = [False] * len(all_input_file_contents)
    input_file_idx = 0
    with open(file_path_prefix, 'w') as combined_file:
        while True:
            if finished_reading[input_file_idx]:
                input_file_idx = (input_file_idx + 1) % len(all_input_file_contents)
                continue
            for i in range(line_group_size):
                line = all_input_file_contents[input_file_idx][all_indices[input_file_idx]]
                combined_file.write(line)
                all_indices[input_file_idx] += 1
            if all_indices[input_file_idx] == len(all_input_file_contents[input_file_idx]):
                finished_reading[input_file_idx] = True
                if all(finished_reading):
                    break
            input_file_idx = (input_file_idx + 1) % len(all_input_file_contents)

    if delete:
        for file_path in all_input_file_paths:
            os.remove(file_path)


def map_filter(callable, iterable):
    output = []
    for element in iterable:
        new_element = callable(element)
        if new_element is not None:
            output.append(new_element)
    return output


def get_devices(devices=None):
    if not torch.cuda.is_available():
        return [torch.device('cpu')]
    if not devices:
        return [torch.device('cuda:' + str(i)) for i in range(torch.cuda.device_count())]
    return [torch.device(ordinal) for ordinal in devices]


def set_seed(args):
    np.random.seed(args.seed)
    random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed_all(args.seed)


def get_trainable_params(model, name=False):
    # TODO is always called with name=False, so remove the if statement
    if name:
        return list(filter(lambda p: p[1].requires_grad, model.named_parameters()))
    else:
        return list(filter(lambda p: p.requires_grad, model.parameters()))


def log_model_size(logger, model, model_name):
    num_param = sum([p.nelement() for p in model.parameters() if p.requires_grad])
    logger.info(f'{model_name} has {num_param:,} parameters')


def elapsed_time(log):
    t = time.time() - log.start
    day = int(t // (24 * 3600))
    t = t % (24 * 3600)
    hour = int(t // 3600)
    t %= 3600
    minutes = int(t // 60)
    t %= 60
    seconds = int(t)
    return f'{day:02}:{hour:02}:{minutes:02}:{seconds:02}'


def make_data_loader(dataset, numericalizer, batch_size, device=None, train=False, return_original_order=False):
    all_features = NumericalizedExamples.from_examples(dataset, numericalizer=numericalizer)

    context_lengths = [ex.context.length for ex in all_features]
    answer_lengths = [ex.answer.length for ex in all_features]

    logger.info(
        f'context lengths (min, mean, max): {np.min(context_lengths)}, {int(np.mean(context_lengths))}, {np.max(context_lengths)}'
    )
    logger.info(
        f'answer lengths (min, mean, max): {np.min(answer_lengths)}, {int(np.mean(answer_lengths))}, {np.max(answer_lengths)}'
    )

    sampler = LengthSortedIterator(
        all_features,
        batch_size=batch_size,
        sort=True,
        shuffle_and_repeat=train,
        sort_key_fn=dataset.sort_key_fn,
        batch_size_fn=dataset.batch_size_fn,
        groups=dataset.groups,
    )
    # get the sorted data_source
    all_f = sampler.data_source
    data_loader = torch.utils.data.DataLoader(
        all_f,
        batch_sampler=sampler,
        collate_fn=lambda batches: NumericalizedExamples.collate_batches(batches, numericalizer, device),
        num_workers=0,
    )

    if return_original_order:
        return data_loader, sampler.original_order
    else:
        return data_loader


def ned_dump_entity_type_pairs(dataset, path, name, utterance_field):

    with open(os.path.join(path, f'{name}_labels.jsonl'), 'w') as fout:
        for ex in dataset.examples:
            text = ex.question_plus_types if utterance_field == 'question' else ex.context_plus_types

            span_beg = text.index('<e>')
            sentence = text[:span_beg].strip()
            entity_span = text[span_beg:].strip()
            entity_token_string = entity_span[len('<e>') : -len('</e>')].strip('; ')

            entities = []
            ent_types = []
            if entity_token_string:
                entity_token_pairs = entity_token_string.split(';')
                for str in entity_token_pairs:
                    entity, types = token_type_regex.match(str).groups()
                    types = types.split('|')
                    entities.append(entity.strip())
                    ent_types.append(types)

            fout.write(ujson.dumps({"sentence": sentence, "aliases": entities, "thingtalk_types": ent_types}) + '\n')


def get_mbart_lang(orig_lang):
    for lang in FAIRSEQ_LANGUAGE_CODES:
        if lang.startswith(orig_lang):
            return lang


def adjust_language_code(config, pretrained_model, src_lang, tgt_lang):

    # adjust src and tgt languages for Marian models
    model_is_marian = isinstance(config, MarianConfig)

    if model_is_marian and pretrained_model.rsplit('-', 2)[1] in MARIAN_GROUP_MEMBERS:
        if src_lang not in MARIAN_GROUP_MEMBERS[pretrained_model.rsplit('-', 2)[1]]:
            if src_lang == 'pl':
                src_lang = 'pol'
            elif src_lang == 'fa':
                src_lang = 'pes'
            else:
                raise ValueError(
                    'Source language is not in this Marian model group languages, please specify the correct source language.'
                )

    if model_is_marian and pretrained_model.rsplit('-', 1)[1] in MARIAN_GROUP_MEMBERS:
        if tgt_lang not in MARIAN_GROUP_MEMBERS[pretrained_model.rsplit('-', 1)[1]]:
            if tgt_lang == 'pl':
                tgt_lang = 'pol'
            elif tgt_lang == 'fa':
                tgt_lang = 'pes'
            else:
                raise ValueError(
                    'Target language is not in this Marian model group languages, please specify the correct target language.'
                )

    if model_is_marian and pretrained_model.rsplit('-', 2)[1] not in MARIAN_GROUP_MEMBERS:
        # Source language should not be provided when using marian models with single language pairs
        # otherwise the translation outputs will be incorrect; hence we ignore the source language
        src_lang = None

    if model_is_marian and pretrained_model.rsplit('-', 1)[1] not in MARIAN_GROUP_MEMBERS:
        # Target language should not be provided when using marian models with single language pairs
        # otherwise the translation outputs will be incorrect; hence we ignore the target language
        tgt_lang = None

    # adjust src and tgt languages for Mbart models
    if isinstance(config, MBartConfig):
        src_lang = get_mbart_lang(src_lang)
        tgt_lang = get_mbart_lang(tgt_lang)

    return src_lang, tgt_lang


def have_multilingual(task_names):
    return any(['multilingual' in name for name in task_names])


def load_config_json(args):
    args.almond_type_embeddings = False
    with open(os.path.join(args.path, 'config.json')) as config_file:
        config = json.load(config_file)

        retrieve = [
            'model',
            'pretrained_model',
            'rnn_dimension',
            'rnn_layers',
            'rnn_zero_state',
            'max_generative_vocab',
            'lower',
            'trainable_decoder_embeddings',
            'override_context',
            'override_question',
            'almond_lang_as_question',
            'almond_has_multiple_programs',
            'almond_detokenize_sentence',
            'almond_thingtalk_version',
            'preprocess_special_tokens',
            'dropper_ratio',
            'dropper_min_count',
            'label_smoothing',
            'use_encoder_loss',
            'num_workers',
            'no_fast_tokenizer',
            'force_fast_tokenizer',
            'override_question',
            'override_context',
            'add_types_to_text',
            'do_ned',
            'database_type',
            'min_entity_len',
            'max_entity_len',
            'entity_type_agg_method',
            'entity_word_embeds_dropout',
            'num_db_types',
            'db_unk_id',
            'ned_retrieve_method',
            'database_lookup_method',
            'almond_domains',
            'ned_features',
            'ned_features_size',
            'ned_features_default_val',
            'bootleg_output_dir',
            'bootleg_model',
            'bootleg_prob_threshold',
            'bootleg_post_process_types',
            'att_pooling',
            'plot_heatmaps',
            'replace_qp',
            'force_replace_qp',
            'no_separator',
            'num_labels',
            'ner_domains',
            'hf_test_overfit',
        ]

        # train and predict scripts have these arguments in common. We use the values from train only if they are not provided in predict
        overwrite = [
            'val_batch_size',
            'num_beams',
            'num_beam_groups',
            'diversity_penalty',
            'num_outputs',
            'no_repeat_ngram_size',
            'top_p',
            'top_k',
            'repetition_penalty',
            'temperature',
            'max_output_length',
            'reduce_metrics',
            'database_dir',
        ]
        for o in overwrite:
            if o not in args or getattr(args, o) is None:
                retrieve.append(o)

        for r in retrieve:
            if r in config:
                setattr(args, r, config[r])
            # These are for backward compatibility with models that were trained before we added these arguments
            elif r in (
                'do_ned',
                'use_encoder_loss',
                'almond_has_multiple_programs',
                'almond_lang_as_question',
                'preprocess_special_tokens',
                'almond_thingtalk_version',
                'plot_heatmaps',
                'replace_qp',
                'force_replace_qp',
                'no_fast_tokenizer',
                'force_fast_tokenizer',
            ):
                setattr(args, r, False)
            elif r in ('num_db_types', 'db_unk_id', 'num_workers'):
                setattr(args, r, 0)
            elif r in ('entity_word_embeds_dropout'):
                setattr(args, r, 0.0)
            elif r in ('num_beams', 'num_outputs', 'top_p', 'repetition_penalty'):
                setattr(args, r, [1])
            elif r in ('no_repeat_ngram_size', 'top_k', 'temperature'):
                setattr(args, r, [0])
            elif r in ['ned_features', 'ned_features_size', 'ned_features_default_val']:
                setattr(args, r, [])
            elif r == 'add_types_to_text':
                setattr(args, r, 'no')
            elif r == 'database_type':
                setattr(args, r, 'json')
            elif r == 'att_pooling':
                setattr(args, r, 'max')
            elif r == 'min_entity_len':
                setattr(args, r, 2)
            elif r == 'max_entity_len':
                setattr(args, r, 4)
            elif r == 'database_dir':
                setattr(args, r, None)
            elif r == 'ned_retrieve_method':
                setattr(args, r, 'naive')
            elif r == 'database_lookup_method':
                setattr(args, r, 'ngrams')
            elif r == 'almond_domains':
                setattr(args, r, [])
            elif r == 'locale':
                setattr(args, r, 'en')
            elif r == 'num_beam_groups':
                setattr(args, r, [1])
            elif r == 'diversity_penalty':
                setattr(args, r, [0.0])
            elif r == 'dropper_ratio':
                setattr(args, r, 0.0)
            elif r == 'dropper_min_count':
                setattr(args, r, 10000)
            elif r == 'label_smoothing':
                setattr(args, r, 0.0)
            elif r == 'no_separator':
                setattr(args, r, True)  # old models don't use a separator
            else:
                # use default value
                setattr(args, r, None)

        args.dropout_ratio = 0.0
        args.verbose = False

    args.best_checkpoint = os.path.join(args.path, args.checkpoint_name)



#
# Copyright (c) 2021, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

from ..data_utils.example import Example
from ..tasks.base_task import BaseTask
from ..tasks.registry import register_task
from .hf_dataset import HFDataset

logger = logging.getLogger(__name__)


class HFTask(BaseTask):
    def __init__(self, name, args):
        super().__init__(name, args)

    def utterance_field(self):
        return 'question'


@register_task('ambig_qa')
class AmbigQA(HFTask):
    def __init__(self, name, args):
        super().__init__(name, args)

    @property
    def metrics(self):
        return ['em', 'bleu']

    def _make_example(self, ex, **kwargs):
        example_id, question, used_queries, annotations, nq_answer = (
            ex['id'],
            ex['question'],
            ex['used_queries'],
            ex['annotations'],
            ex['nq_answer'],
        )

        # assert len(nq_answer) == 1, print(example_id, nq_answer)

        # TODO choosing only first answer for now
        answer = nq_answer[0]

        # process used_queries
        # assert len(used_queries['results']) == 1, print(example_id, used_queries)
        # TODO choosing only first used_query for now
        context = used_queries['query'][0] + ' <q> ' + ' <p> '.join(used_queries['results'][0]['snippet'])
        context_tokens = context.split(' ')[:180]
        context_tokens += ['<u>']
        context = ' '.join(context_tokens)

        if annotations['type'] == 'singleAnswer':
            # unambiguous question
            assert len(annotations['answer']) == 1
        else:
            # find the disambiguated question with the correct nq answer
            all_answers = annotations['qaPairs'][0]['answer']
            all_questions = annotations['qaPairs'][0]['question']
            assert len(all_answers) == len(all_questions)
            for q, a in zip(all_questions, all_answers):
                if a == nq_answer:
                    question = q

        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def get_splits(self, root, **kwargs):
        return HFDataset.return_splits(name=self.name, path=root, make_example=self._make_example, **kwargs)


@register_task('conll2003')
class CONLLNER(HFTask):
    tagging_scheme = 'IOB2'
    conll_labels = ['O', 'B-PER', 'I-PER', 'B-ORG', 'I-ORG', 'B-LOC', 'I-LOC', 'B-MISC', 'I-MISC']
    domain2labels = {"conll": conll_labels}

    def __init__(self, name, args):
        self.all_labels = tuple(self.conll_labels)
        self.label2id = {}
        for label in self.all_labels:
            if label not in self.label2id:
                self.label2id[label] = len(self.label2id)
        self.id2label = {v: k for k, v in self.label2id.items()}
        self.num_labels = len(self.label2id)
        super().__init__(name, args)

    @property
    def metrics(self):
        return ['ner_f1', 'em', 'f1', 'pem']

    def utterance_field(self):
        return 'context'

    def _make_example(self, ex, **kwargs):
        example_id = ex['id']
        context = ' '.join(ex['tokens'])
        question = ''
        answer = ' '.join(map(lambda item: str(item), ex['ner_tags']))

        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def get_splits(self, root, **kwargs):
        splits, paths = HFDataset.return_splits(name=self.name, path=root, make_example=self._make_example, **kwargs)
        for split in splits:
            if split:
                split.is_classification = True
        return splits, paths
